import requests


def get_exception_from_response(response: requests.Response) -> Exception:
    text = response.content.decode()
    if text == "internal-error" or response.status_code == 500:
        return ActualError(text)
    # taken from
    # https://github.com/actualbudget/actual-server/blob/6e9eddeb561b0d9f2bbb6301c3e2c30b4effc522/src/app-sync.js#L107
    elif text == "file-has-new-key":
        return ActualError(f"{text}: The data is encrypted with a different key")
    elif text == "file-has-reset":
        return InvalidFile(
            f"{text}: The changes being synced are part of an old group, which means the file has been reset. "
            f"User needs to re-download."
        )
    elif text in ("file-not-found", "file-needs-upload"):
        raise UnknownFileId(text)
    elif text == "file-old-version":
        raise InvalidFile(f"{text}: SYNC_FORMAT_VERSION was generated with an old format")


class ActualError(Exception):
    """General error with Actual. The error message should provide more information."""

    pass


class ActualInvalidOperationError(ActualError):
    """Invalid operation requested. Happens usually when a request has been done, but it's missing a required
    parameters."""

    pass


class AuthorizationError(ActualError):
    """When the login fails due to invalid credentials, or a request has been done with the wrong credentials
    (i.e. invalid token)"""

    pass


class UnknownFileId(ActualError):
    """When the file id that has been set does not exist on the server."""

    pass


class InvalidZipFile(ActualError):
    """
    The validation fails when loading a zip file, either because it's an invalid zip file or the file is corrupted.
    """

    pass


class InvalidFile(ActualError):

    pass


class ActualDecryptionError(ActualError):
    """
    The decryption for the file failed. This can happen for a multitude or reasons, like the password is wrong, the file
    is corrupted, or when the password is not provided but the file is encrypted.
    """

    pass


class ActualSplitTransactionError(ActualError):
    """The split transaction is invalid, most likely because the sum of splits is not equal the full amount of the
    transaction."""

    pass


class ActualBankSyncError(ActualError):
    """The bank sync had an error, due to the service being unavailable or due to authentication issues with the
    third-party service. This likely indicates a problem with the configuration of the bank sync, not an issue with
    this library."""

    def __init__(self, error_type: str, status: str = None, reason: str = None):
        self.error_type, self.status, self.reason = error_type, status, reason



from __future__ import annotations

import datetime
import decimal
import json
import typing
import uuid
import warnings

import sqlalchemy
from pydantic import TypeAdapter
from sqlalchemy import func
from sqlalchemy.orm import joinedload
from sqlalchemy.sql.expression import Select
from sqlmodel import Session, select

from actual.crypto import is_uuid
from actual.database import (
    Accounts,
    Categories,
    CategoryGroups,
    CategoryMapping,
    MessagesClock,
    PayeeMapping,
    Payees,
    Preferences,
    ReflectBudgets,
    Rules,
    Schedules,
    Transactions,
    ZeroBudgets,
)
from actual.exceptions import ActualError
from actual.protobuf_models import HULC_Client
from actual.rules import Action, Condition, Rule, RuleSet
from actual.utils.conversions import cents_to_decimal, current_timestamp, date_to_int, decimal_to_cents, month_range
from actual.utils.title import title

T = typing.TypeVar("T")


def _transactions_base_query(
    s: Session,
    start_date: datetime.date = None,
    end_date: datetime.date = None,
    account: Accounts | str | None = None,
    category: Categories | str | None = None,
    include_deleted: bool = False,
) -> Select:
    query = (
        select(Transactions)
        .options(
            joinedload(Transactions.account),
            joinedload(Transactions.category),
            joinedload(Transactions.payee),
        )
        .filter(
            Transactions.date.isnot(None),
            Transactions.acct.isnot(None),
        )
        .order_by(
            Transactions.date.desc(),
            Transactions.starting_balance_flag,
            Transactions.sort_order.desc(),
            Transactions.id,
        )
    )
    if start_date:
        query = query.filter(Transactions.date >= date_to_int(start_date))
    if end_date:
        query = query.filter(Transactions.date < date_to_int(end_date))
    if not include_deleted:
        query = query.filter(sqlalchemy.func.coalesce(Transactions.tombstone, 0) == 0)
    if account:
        account = get_account(s, account)
        if account:
            query = query.filter(Transactions.acct == account.id)
    if category:
        category = get_category(s, category)
        if category:
            query = query.filter(Transactions.category_id == category.id)
    return query


def _balance_base_query(
    s: Session,
    start_date: datetime.date,
    end_date: datetime.date,
    account: Accounts | str | None = None,
    category: Categories | str | None = None,
) -> Select:
    query = select(func.coalesce(func.sum(Transactions.amount), 0)).where(
        Transactions.date >= date_to_int(start_date),
        Transactions.date < date_to_int(end_date),
        Transactions.is_parent == 0,
        Transactions.tombstone == 0,
    )
    if account:
        account = get_account(s, account)
        if account:
            query = query.filter(Transactions.acct == account.id)
    if category:
        category = get_category(s, category)
        if category:
            query = query.filter(Transactions.category_id == category.id)
    return query


def get_transactions(
    s: Session,
    start_date: datetime.date = None,
    end_date: datetime.date = None,
    notes: str = None,
    account: Accounts | str | None = None,
    category: Categories | str | None = None,
    is_parent: bool = False,
    include_deleted: bool = False,
    budget: ZeroBudgets | None = None,
) -> typing.Sequence[Transactions]:
    """
    Returns a list of all available transactions, sorted by date in descending order.

    :param s: session from Actual local database.
    :param start_date: optional start date for the transaction period (inclusive)
    :param end_date: optional end date for the transaction period (exclusive)
    :param notes: optional notes filter for the transactions. This looks for a case-insensitive pattern rather than for
    the exact match, i.e. 'foo' would match 'Foo Bar'.
    :param account: optional account (either Account object or Account name) filter for the transactions.
    :param category: optional category (either Category object or Category name) filter for the transactions.
    :param is_parent: optional boolean flag to indicate if a transaction is a parent. Parent transactions are either
    single transactions or the main transaction with `Transactions.splits` property. Default is to return all individual
    splits, and the parent can be retrieved by `Transactions.parent`.
    :param include_deleted: includes deleted transactions from the search.
    :param budget: optional budget filter for the transactions. The budget range and category will be used to filter the
                   final results. **Usually not used together with the `start_date` and `end_date` filters, as they
                   might hide results.
    :return: list of transactions with `account`, `category` and `payee` preloaded.
    """
    query = _transactions_base_query(s, start_date, end_date, account, category, include_deleted)
    query = query.filter(Transactions.is_parent == int(is_parent))
    if notes:
        query = query.filter(Transactions.notes.ilike(f"%{sqlalchemy.text(notes).compile()}%"))
    if budget:
        budget_start, budget_end = budget.range
        if (start_date and start_date >= budget_end) or (end_date and end_date < budget_start):
            warnings.warn(
                f"Provided date filters [{start_date}, {end_date}) to get_transactions are outside the bounds of the "
                f"budget range [{budget_start}, {budget_end}). Results might be empty!"
            )
        budget_start, budget_end = (date_to_int(d) for d in budget.range)
        query = query.filter(
            Transactions.date >= budget_start,
            Transactions.date < budget_end,
            Transactions.category_id == budget.category_id,
        )
    return s.exec(query).all()


def match_transaction(
    s: Session,
    date: datetime.date,
    account: str | Accounts,
    payee: str | Payees = "",
    amount: decimal.Decimal | float | int = 0,
    imported_id: str | None = None,
    already_matched: typing.List[Transactions] = None,
) -> typing.Optional[Transactions]:
    """Matches a transaction with another transaction based on the fuzzy matching described at
    [`reconcileTransactions`](
    https://github.com/actualbudget/actual/blob/b192ad955ed222d9aa388fe36557b39868029db4/packages/loot-core/src/server/accounts/sync.ts#L347)

    The matches, from strongest to the weakest are defined as follows:

    - The strongest match will be the imported_id (or financial_id)
    - The transaction with the same exact amount and around the same date (7 days), with the same payee (closest first)
    - The transaction with the same exact amount and around the same date (7 days, closest first)

    """
    # First, match with an existing transaction's imported_id
    if imported_id:
        query = _transactions_base_query(s, account=account)
        imported_transaction = s.exec(query.filter(Transactions.financial_id == imported_id)).first()
        if imported_transaction:
            return imported_transaction  # noqa
    # if not matched, look 7 days ahead and 7 days back when fuzzy matching
    query = _transactions_base_query(
        s, date - datetime.timedelta(days=7), date + datetime.timedelta(days=8), account=account
    ).filter(Transactions.amount == round(amount * 100))
    results: typing.List[Transactions] = s.exec(query).all()  # noqa
    # filter out the ones that were already matched
    if already_matched:
        matched = {t.id for t in already_matched}
        results = [r for r in results if r.id not in matched]
    if not results:
        # nothing to be matched
        return None
    # sort the results by their distance to the original date
    results.sort(key=lambda t: abs((t.get_date() - date).total_seconds()))
    # Next, do the fuzzy matching. This first pass matches based on the
    # payee id. We do this in multiple passes so that higher fidelity
    # matching always happens first, i.e. a transaction should
    # match with low fidelity if a later transaction is going to match
    # the same one with high fidelity.
    payee = get_payee(s, payee)
    if payee:
        matching_payee = [r for r in results if r.payee_id == payee.id]
        if matching_payee:
            return matching_payee[0]
    # The final fuzzy matching pass. This is the lowest fidelity
    # matching: it just find the first transaction that hasn't been
    # matched yet. Remember the dataset only contains transactions
    # around the same date with the same amount.
    return results[0]


def create_transaction_from_ids(
    s: Session,
    date: datetime.date,
    account_id: str,
    payee_id: typing.Optional[str],
    notes: str,
    category_id: str = None,
    amount: decimal.Decimal = 0,
    imported_id: str = None,
    cleared: bool = False,
    imported_payee: str = None,
    process_payee: bool = True,
) -> Transactions:
    """Internal method to generate a transaction from ids instead of objects."""
    date_int = date_to_int(date)
    t = Transactions(
        id=str(uuid.uuid4()),
        acct=account_id,
        date=date_int,
        amount=decimal_to_cents(amount),
        category_id=category_id,
        notes=notes,
        reconciled=0,
        cleared=int(cleared),
        sort_order=current_timestamp(),
        financial_id=imported_id,
        imported_description=imported_payee,
    )
    s.add(t)
    if process_payee:
        set_transaction_payee(s, t, payee_id)
    return t


def create_transaction(
    s: Session,
    date: datetime.date,
    account: str | Accounts,
    payee: str | Payees | None = None,
    notes: str | None = "",
    category: str | Categories | None = None,
    amount: decimal.Decimal | float | int = 0,
    imported_id: str | None = None,
    cleared: bool = False,
    imported_payee: str = None,
) -> Transactions:
    """
    Creates a transaction from the provided input.

    :param s: session from Actual local database.
    :param date: date of the transaction.
    :param account: either account name or account object (via `get_account` or `get_accounts`). Will not be
    auto-created if missing.
    :param payee: optional name of the payee from the transaction. Will be created if missing.
    :param notes: optional description for the transaction.
    :param category: optional category for the transaction. Will be created if not existing.
    :param amount: amount of the transaction. Positive indicates that the account balance will go up (deposit), and
    negative that the account balance will go down (payment)
    :param imported_id: unique id of the imported transaction. This is often provided if the transaction comes from
    a third-party system that contains unique ids (i.e. via bank sync).
    :param cleared: visual indication that the transaction is in both your budget and in your account statement,
    and they match.
    :param imported_payee: known internally as imported_description, this is the original name of the payee, when
    importing data and before running rules.
    :return: the generated transaction object.
    """
    acct = get_account(s, account)
    if acct is None:
        raise ActualError(f"Account {account} not found")
    if imported_payee:
        imported_payee = imported_payee.strip()
        if not payee:
            payee = imported_payee
    payee_id = None
    if payee is not None:
        payee_id = get_or_create_payee(s, payee).id
    if category:
        category_id = get_or_create_category(s, category).id
    else:
        category_id = None

    return create_transaction_from_ids(
        s, date, acct.id, payee_id, notes, category_id, amount, imported_id, cleared, imported_payee
    )


def set_transaction_payee(s: Session, transaction: Transactions, payee: typing.Union[Payees, str, None]) -> None:
    """
    Sets a payee safely by checking if this payee is a transfer. If it is, then the transfer will be created.

    This is necessary since the payee can be set to a "transfer id", which references and account. When this happens,
    the transaction will be marked as a transfer between the two accounts, and new transaction will need to be created
    on the other account, with the negative amount.

    :param s: session from Actual local database.
    :param transaction: transaction to exchange the payee.
    :param payee: object or unique id of the payee to be set. Must be existing
    """
    current_payee: typing.Optional[Payees] = None
    if isinstance(payee, str):
        payee = s.scalar(select(Payees).where(Payees.id == payee))
    if transaction.payee_id:  # resolve based on the id, in case the relationship did not load
        current_payee = s.scalar(select(Payees).where(Payees.id == transaction.payee_id))
    # if old payee was a transfer, we delete that transfer based on the `transferred_id`
    if current_payee and current_payee.transfer_acct:
        old_tr = s.scalar(select(Transactions).where(Transactions.id == transaction.transferred_id))
        if old_tr:  # should exist, but check for safety
            old_tr.delete()
            transaction.transferred_id = None
    # if setting a transfer payee, we create a transfer
    if payee and payee.transfer_acct:
        transfer = create_transaction_from_ids(
            s,
            transaction.get_date(),
            payee.account.id,
            transaction.account.payee.id,
            transaction.notes,
            None,
            cleared=bool(transaction.cleared),
            amount=-transaction.get_amount(),
            process_payee=False,
        )
        transaction.category_id = None
        transfer.transferred_id, transaction.transferred_id = transaction.id, transfer.id

    # finally set the payee
    transaction.payee_id = payee.id if payee else None


def normalize_payee(payee_name: str | None, raw_payee_name: bool = False) -> str:
    """
    Normalizes the payees according to the source code found at the [official source code](
    https://github.com/actualbudget/actual/blob/f02ca4e3d26f5b91f4234317e024022fcae2c13c/packages/loot-core/src/server/accounts/sync.ts#L206-L214)

    This make sures that the payees are consistent across the imports, i.e. 'MY PAYEE ' turns into 'My Payee', but so
    does 'My PaYeE'.

    :param payee_name: the original payee name to be normalized.
    :param raw_payee_name: if the original payee name should be used instead. If the payee provided consists of spaces
    or an empty string, it will still be assigned to `None`.
    :return:
    """
    if payee_name:
        trimmed = payee_name.strip()
        if raw_payee_name:
            return trimmed
        else:
            return title(trimmed)
    return ""


def reconcile_transaction(
    s: Session,
    date: datetime.date,
    account: str | Accounts,
    payee: str | Payees = "",
    notes: str = "",
    category: str | Categories | None = None,
    amount: decimal.Decimal | float | int = 0,
    imported_id: str | None = None,
    cleared: bool = False,
    imported_payee: str = None,
    update_existing: bool = True,
    already_matched: typing.List[Transactions] = None,
) -> Transactions:
    """Matches the transaction to an existing transaction using fuzzy matching.

    :param s: session from Actual local database.
    :param date: date of the transaction.
    :param account: either account name or account object (via `get_account` or `get_accounts`). Will not be
    auto-created if missing.
    :param payee: name of the payee from the transaction. Will be created if missing.
    :param notes: optional description for the transaction.
    :param category: optional category for the transaction. Will be created if not existing.
    :param amount: amount of the transaction. Positive indicates that the account balance will go up (deposit), and
    negative that the account balance will go down (payment)
    :param imported_id: unique id of the imported transaction. This is often provided if the transaction comes from
    a third-party system that contains unique ids (i.e. via bank sync).
    :param cleared: This is a visual indication that the transaction is in both your budget and in your account
    statement, and they match.
    :param imported_payee: known internally as imported_description, this is the original name of the payee, when
    importing data and before running rules.
    :param update_existing: if the transaction should be updated to the provided properties, if a match is found.
    :param already_matched: list of the transactions that were already matched. When importing a list of transactions,
    this would prevent transactions with the exact same (date, amount) to be assigned as duplicates.
    :return: the generated or matched transaction object.
    """
    account = get_account(s, account)
    match = match_transaction(s, date, account, payee, amount, imported_id, already_matched)
    if match:
        # try to update fields
        if update_existing:
            match.notes = notes
            if category:
                match.category_id = get_or_create_category(s, category).id
            match.set_date(date)
        return match
    return create_transaction(s, date, account, payee, notes, category, amount, imported_id, cleared, imported_payee)


def create_splits(
    s: Session, transactions: typing.Sequence[Transactions], payee: str | Payees = "", notes: str = ""
) -> Transactions:
    """
    Creates a transaction with splits based on the list of transactions. The total amount will be evaluated as the sum
    of the individual amounts. All dates must be set to the same value.

    :param s: session from Actual local database.
    :param transactions: list of transactions that will be added to the splits.
    :param payee: name or object of the payee from the transaction. Will be created if missing.
    :param notes: optional description for the transaction.
    :return: the generated transaction object for the parent transaction.
    """
    if not all(transactions[0].date == t.date for t in transactions) or not all(
        transactions[0].acct == t.acct for t in transactions
    ):
        raise ActualError("`date` and `acct` must be the same for all transactions in splits")
    payee = get_or_create_payee(s, payee)
    split_amount = decimal.Decimal(sum(t.get_amount() for t in transactions))
    split_transaction = create_transaction_from_ids(
        s, transactions[0].get_date(), transactions[0].acct, payee.id, notes, None, split_amount
    )
    split_transaction.is_parent = 1
    split_transaction.is_child = 0
    for transaction in transactions:
        transaction.is_parent = 0
        transaction.is_child = 1
        transaction.parent_id = split_transaction.id
    return split_transaction


def create_split(s: Session, transaction: Transactions, amount: float | decimal.Decimal) -> Transactions:
    """
    Creates a transaction split based on the parent transaction. This is the opposite of create_splits, that joins
    all transactions as one big transaction. When using this method, you need to make sure all splits that you add to
    a transaction are then valid.

    :param s: session from Actual local database.
    :param transaction: parent transaction to the split you want to create.
    :param amount: amount of the split.
    :return: the generated transaction object for the split transaction.
    """
    split = create_transaction(
        s, transaction.get_date(), transaction.account, transaction.payee, None, transaction.category, amount=amount
    )
    split.parent_id, split.is_parent, split.is_child = transaction.id, 0, 1
    return split


def base_query(instance: typing.Type[T], name: str = None, include_deleted: bool = False) -> Select:
    """Internal method to reduce querying complexity on sub-functions."""
    query = select(instance)
    if not include_deleted:
        query = query.filter(sqlalchemy.func.coalesce(instance.tombstone, 0) == 0)
    if name:
        query = query.filter(instance.name.ilike(f"%{sqlalchemy.text(name).compile()}%"))
    return query


def create_category_group(s: Session, name: str) -> CategoryGroups:
    """Creates a new category with the group name `name`. Make sure you avoid creating payees with duplicate names, as
    it makes it difficult to find them without knowing the unique id beforehand."""
    category_group = CategoryGroups(id=str(uuid.uuid4()), name=name, is_income=0, sort_order=0)
    s.add(category_group)
    return category_group


def get_or_create_category_group(s: Session, name: str) -> CategoryGroups:
    """Gets or create the category group, if not found with `name`. Deleted category groups are excluded from the
    search."""
    category_group = s.exec(
        select(CategoryGroups).filter(CategoryGroups.name == name, CategoryGroups.tombstone == 0)
    ).one_or_none()
    if not category_group:
        category_group = create_category_group(s, name)
    return category_group


def get_categories(s: Session, name: str = None, include_deleted: bool = False) -> typing.Sequence[Categories]:
    """
    Returns a list of all available categories.

    :param s: session from Actual local database.
    :param name: pattern name of the payee, case-insensitive.
    :param include_deleted: includes all payees which were deleted via frontend. They would not show normally.
    :return: list of categories with `transactions` already loaded.
    """
    query = base_query(Categories, name, include_deleted).options(joinedload(Categories.transactions))
    return s.exec(query).unique().all()


def create_category(
    s: Session,
    name: str,
    group_name: str = None,
) -> Categories:
    """Creates a new category with the `name` and `group_name`. If the group is not existing, it will also be created.
    Make sure you avoid creating categories with duplicate names, as it makes it difficult to find them without knowing
    the unique id beforehand. The exception is to have them in separate group names, but you then need to provide the
    group name to the method also.

    If a group name is not provided, the default 'Usual Expenses' will be picked.
    """
    category_group = get_or_create_category_group(s, group_name if group_name is not None else "Usual Expenses")
    category = Categories(
        id=str(uuid.uuid4()), name=name, hidden=0, is_income=0, sort_order=0, cat_group=category_group.id
    )
    category_mapping = CategoryMapping(id=category.id, transfer_id=category.id)
    s.add(category)
    s.add(category_mapping)
    return category


def get_category(
    s: Session, name: str | Categories, group_name: str = None, strict_group: bool = False
) -> typing.Optional[Categories]:
    """Gets an existing category by name, returns `None` if not found. Deleted payees are excluded from the search."""
    if isinstance(name, Categories):
        return name
    category = s.exec(
        select(Categories)
        .join(CategoryGroups)
        .filter(Categories.name == name, Categories.tombstone == 0, CategoryGroups.name == group_name)
    ).one_or_none()
    if not category and not strict_group:
        # try to find it without the group name
        category = s.exec(select(Categories).filter(Categories.name == name, Categories.tombstone == 0)).one_or_none()
    return category


def get_or_create_category(
    s: Session, name: str | Categories, group_name: str = None, strict_group: bool = False
) -> Categories:
    """Gets or create the category, if not found with `name`. If the category already exists, but in a different group,
    but the category name is still unique, it will be returned, unless `strict_group` is set to `True`.

    If a group name is not provided, the default 'Usual Expenses' will be picked.
    """
    category = get_category(s, name, group_name, strict_group)
    if not category:
        category = create_category(s, name, group_name or "Usual Expenses")
    return category


def get_accounts(s: Session, name: str = None, include_deleted: bool = False) -> typing.Sequence[Accounts]:
    """
    Returns a list of all available accounts.

    :param s: session from Actual local database.
    :param name: pattern name of the payee, case-insensitive.
    :param include_deleted: includes all payees which were deleted via frontend. They would not show normally.
    :return: list of accounts with `transactions` already loaded.
    """
    query = base_query(Accounts, name, include_deleted).options(joinedload(Accounts.transactions))
    return s.exec(query).unique().all()


def get_payees(s: Session, name: str = None, include_deleted: bool = False) -> typing.Sequence[Payees]:
    """
    Returns a list of all available payees.

    :param s: session from Actual local database.
    :param name: pattern name of the payee, case-insensitive.
    :param include_deleted: includes all payees which were deleted via frontend. They would not show normally.
    :return: list of payees with `transactions` already loaded.
    """
    query = base_query(Payees, name, include_deleted).options(joinedload(Payees.transactions))
    return s.exec(query).unique().all()


def get_payee(s: Session, name: str | Payees) -> typing.Optional[Payees]:
    """Gets an existing payee by name, returns `None` if not found. Deleted payees are excluded from the search."""
    if isinstance(name, Payees):
        return name
    return s.exec(select(Payees).filter(Payees.name == name, Payees.tombstone == 0)).one_or_none()


def create_payee(s: Session, name: str | None) -> Payees:
    """Creates a new payee with the desired name. Make sure you avoid creating payees with duplicate names, as it makes
    it difficult to find them without knowing the unique id beforehand."""
    payee = Payees(id=str(uuid.uuid4()), name=name)
    s.add(payee)
    # add also the payee mapping
    s.add(PayeeMapping(id=payee.id, target_id=payee.id))
    return payee


def get_or_create_payee(s: Session, name: str | Payees) -> Payees:
    """Gets an existing payee by name, and if it does not exist, creates a new one. If the payee is created twice,
    this method will fail with a database error."""
    payee = get_payee(s, name)
    if not payee:
        payee = create_payee(s, name)
    return payee


def create_account(
    s: Session, name: str, initial_balance: decimal.Decimal | float = decimal.Decimal(0), off_budget: bool = False
) -> Accounts:
    """Creates a new account with the name and balance. Make sure you avoid creating accounts with duplicate names, as
    it makes it difficult to find them without knowing the unique id beforehand."""
    acct = Accounts(id=str(uuid.uuid4()), name=name, offbudget=int(off_budget), closed=0)
    s.add(acct)
    # add a blank payee
    payee = create_payee(s, None)
    payee.transfer_acct = acct.id
    s.add(payee)
    # if there is no initial balance, create it
    if initial_balance:
        payee_starting = get_or_create_payee(s, "Starting Balance")
        category = get_or_create_category(s, "Starting Balances", "Income")
        create_transaction_from_ids(
            s, datetime.date.today(), acct.id, payee_starting.id, "", category.id, initial_balance
        )
    return acct


def get_account(s: Session, name: str | Accounts) -> typing.Optional[Accounts]:
    """
    Gets an account with the desired name, otherwise returns `None`. Deleted accounts are excluded from the search.
    """
    if isinstance(name, Accounts):
        return name
    if is_uuid(name):
        query = select(Accounts).filter(Accounts.id == name, Accounts.tombstone == 0)
    else:
        query = select(Accounts).filter(Accounts.name == name, Accounts.tombstone == 0)
    return s.exec(query).one_or_none()


def get_or_create_account(s: Session, name: str | Accounts) -> Accounts:
    """Gets or create the account, if not found with `name`. The initial balance will be set to 0 if an account is
    created using this method."""
    account = get_account(s, name)
    if not account:
        account = create_account(s, name)
    return account


def _get_budget_table(s: Session) -> typing.Type[typing.Union[ReflectBudgets, ZeroBudgets]]:
    """
    Finds out which type of budget the user uses. The types are:

    - Envelope budgeting (default, recommended): `budgetType` is `rollover` or `envelope`, table is ZeroBudgets
    - Tracking budgeting: `budgetType` is `report` or `tracking`, table is `ReflectBudgets`

    :param s: session from Actual local database.
    :return: table object for the budget type, based on the preferences.
    """
    budget_type = get_preference(s, "budgetType")
    if budget_type and budget_type.value in ("report", "tracking"):
        return ReflectBudgets
    else:  # budgetType is rollover or envelope
        return ZeroBudgets


def get_budgets(
    s: Session, month: datetime.date = None, category: str | Categories = None
) -> typing.Sequence[typing.Union[ZeroBudgets, ReflectBudgets]]:
    """
    Returns a list of all available budgets. The object type returned will be either
    ZeroBudgets or ReflectBudgets, depending on the type of budget selected globally. The budget options are:

    - Envelope budgeting (default): ZeroBudgets
    - Tracking budgeting: ReflectBudgets

    :param s: session from Actual local database.
    :param month: month to get budgets for, as a date for that month. Use `datetime.date.today()` if you want the budget
                  for current month
    :param category: category to filter for the budget. By default, the query looks for all budgets.
    :return: list of budgets. It's important to note that budgets will only exist if they are actively set beforehand.
             When the frontend shows a budget as 0.00, it might not be returned by this method.
    """
    table = _get_budget_table(s)
    query = select(table).options(joinedload(table.category)).order_by(table.month.asc())
    if month:
        month_filter = date_to_int(month, month_only=True)
        query = query.filter(table.month == month_filter)
    if category:
        category = get_category(s, category)
        if not category:
            raise ActualError("Category is provided but does not exist.")
        query = query.filter(table.category_id == category.id)
    return s.exec(query).unique().all()


def get_budget(
    s: Session, month: datetime.date, category: str | Categories
) -> typing.Optional[typing.Union[ZeroBudgets, ReflectBudgets]]:
    """
    Gets an existing budget by category name, returns `None` if not found.

    :param s: session from Actual local database.
    :param month: month to get budgets for, as a date for that month. Use `datetime.date.today()` if you want the budget
                  for current month.
    :param category: category to filter for the budget.
    :return: returns the budget matching the month and category. If not found, returns `None`. If the budget is not
             set via frontend, it will show as 0.00, but this function will still return `None`.
    """
    budgets = get_budgets(s, month, category)
    return budgets[0] if budgets else None


def create_budget(
    s: Session, month: datetime.date, category: str | Categories, amount: decimal.Decimal | float | int = 0.0
) -> typing.Union[ZeroBudgets, ReflectBudgets]:
    """
    Gets an existing budget based on the month and category. If it already exists, the amount will be replaced by
    the new amount.

    :param s: session from Actual local database.
    :param month: month to get budgets for, as a date for that month. Use `datetime.date.today()` if you want the budget
                  for current month.
    :param category: category to filter for the budget.
    :param amount: amount for the budget.
    :return: return budget matching the month and category, and assigns the amount to the budget. If not found, creates
             a new budget.
    """
    table = _get_budget_table(s)
    budget = get_budget(s, month, category)
    if budget:
        budget.set_amount(amount)
        return budget
    category = get_category(s, category)
    budget = table(id=str(uuid.uuid4()), category_id=category.id)
    budget.set_date(month)
    budget.set_amount(amount)
    s.add(budget)
    return budget


def get_budgeted_balance(s: Session, month: datetime.date, category: str | Categories) -> decimal.Decimal:
    """
    Returns the budgeted balance as shown by the Actual UI under the category for the individual month. Does not take
    into account previous months.

    :param s: session from Actual local database.
    :param month: month to get budgets for, as a date for that month. Use `datetime.date.today()` if you want the budget
                  for current month.
    :param category:  category to filter for the budget.
    :return: A decimal representing the budget real balance for the category.
    """

    budget = get_budget(s, month, category)
    if not budget:
        # create a temporary budget
        range_start, range_end = month_range(month)
        balance = s.scalar(_balance_base_query(s, range_start, range_end, category=category))
        budget_leftover = cents_to_decimal(balance)
    else:
        budget_leftover = budget.get_amount() + budget.balance  # we can sum because balance is negative
    return budget_leftover


def _get_first_positive_transaction(s: Session, category: Categories) -> typing.Optional[Transactions]:
    """
    Returns the first positive transaction in a certain category. This is used to find the month to start the
    budgeting calculation, since it makes the budget positive.
    """
    query = select(Transactions).where(Transactions.amount > 0, Transactions.category_id == category.id)
    return s.exec(query).first()


def get_accumulated_budgeted_balance(s: Session, month: datetime.date, category: str | Categories) -> decimal.Decimal:
    """
    Returns the budgeted balance as shown by the Actual UI under the category. This is calculated by summing all
    considered budget values and subtracting all transactions for them.

    When using **envelope budget**, this value will accumulate with each consecutive month that your spending is
    greater than your budget. If this value goes under 0.00, your budget is reset for the next month.

    When using **tracking budget**, only the current month is considering for savings, so no previous values will carry
    over.

    :param s: session from Actual local database.
    :param month: month to get budgets for, as a date for that month. Use `datetime.date.today()` if you want the budget
                  for current month.
    :param category:  category to filter for the budget.
    :return: A decimal representing the budget real balance for the category. This is evaluated by adding all
             previous leftover budgets that have a value greater than 0.
    """
    budgets = get_budgets(s, category=category)
    is_tracking_budget = _get_budget_table(s) is ReflectBudgets
    # the first ever budget is the longest we have to look for when searching for the running balance
    # If the budget is set to tracking, the accumulated value will always be the months balance
    if not budgets or is_tracking_budget:
        return get_budgeted_balance(s, month, category)
    first_budget_month = budgets[0].get_date()
    # Get first positive transaction
    first_positive_transaction = _get_first_positive_transaction(s, category)
    first_transaction_month = (
        first_positive_transaction.get_date() if first_positive_transaction else first_budget_month
    )
    # current month is the least of those two dates
    current_month = min(first_budget_month, first_transaction_month)
    accumulated_balance = decimal.Decimal(0)
    while current_month <= month:
        if accumulated_balance < 0:
            accumulated_balance = decimal.Decimal(0)
        current_month_balance = get_budgeted_balance(s, current_month, category)
        accumulated_balance += current_month_balance
        # go to the next month
        current_month = (current_month.replace(day=1) + datetime.timedelta(days=31)).replace(day=1)
    return accumulated_balance


def create_transfer(
    s: Session,
    date: datetime.date,
    source_account: str | Accounts,
    dest_account: str | Accounts,
    amount: decimal.Decimal | int | float,
    notes: str = None,
) -> typing.Tuple[Transactions, Transactions]:
    """
    Creates a transfer of money between two accounts, from `source_account` to `dest_account`. The amount is provided
    as a positive value.

    :param s: session from Actual local database.
    :param date: date of the transfer.
    :param source_account: account that will transfer the money, and reduce its balance.
    :param dest_account: account that will receive the money, and increase its balance.
    :param amount: amount, as a positive decimal, to be transferred.
    :param notes: additional description for the transfer.
    :return: tuple containing both transactions, as one is created per account. The transactions would be
             cross-referenced by their `transferred_id`.
    """
    if amount <= 0:
        raise ActualError("Amount must be a positive value.")
    source: Accounts = get_account(s, source_account)
    dest: Accounts = get_account(s, dest_account)
    source_transaction = create_transaction_from_ids(
        s, date, source.id, dest.payee.id, notes, None, -amount, process_payee=False
    )
    dest_transaction = create_transaction_from_ids(
        s, date, dest.id, source.payee.id, notes, None, amount, process_payee=False
    )
    # swap the transferred ids
    source_transaction.transferred_id = dest_transaction.id
    dest_transaction.transferred_id = source_transaction.id
    # set payees
    source_transaction.payee_id = dest.payee.id
    dest_transaction.payee_id = source.payee.id
    # add and return objects
    s.add(source_transaction)
    s.add(dest_transaction)
    return source_transaction, dest_transaction


def get_rules(s: Session, include_deleted: bool = False) -> typing.Sequence[Rules]:
    """
    Returns a list of all available rules.

    :param s: session from Actual local database.
    :param include_deleted: includes all payees which were deleted via frontend. They would not show normally.
    :return: list of rules.
    """
    return s.exec(base_query(Rules, None, include_deleted)).all()


def get_ruleset(s: Session) -> RuleSet:
    """
    Returns a list of all available rules, but as a rule set that can be used to be applied to existing transactions.

    :param s: session from Actual local database.
    :return: RuleSet object that contains all rules and can be either.
    """
    rule_set = list()
    for rule in get_rules(s):
        conditions = TypeAdapter(typing.List[Condition]).validate_json(rule.conditions)
        actions = TypeAdapter(typing.List[Action]).validate_json(rule.actions)
        rs = Rule(conditions=conditions, operation=rule.conditions_op, actions=actions, stage=rule.stage)  # noqa
        rule_set.append(rs)
    return RuleSet(rules=rule_set)


def create_rule(
    s: Session,
    rule: Rule,
    run_immediately: bool = False,
) -> Rules:
    """
    Creates a rule based on the conditions and actions defined on the input rule. The rule can be ordered to run
    immediately, running the action for all entries that match the conditions on insertion.

    :param s: session from Actual local database.
    :param rule: a constructed rule object from `actual.rules`. The rule format and data types are validated on the
        constructor, **but the data itself is not. Make sure that, if you reference uuids, that they exist.
    :param run_immediately: if the run should run for all transactions on insert, defaults to `False`.
    :return: Rule database object created.
    """
    conditions = json.dumps([c.as_dict() for c in rule.conditions])
    actions = json.dumps([a.as_dict() for a in rule.actions])
    database_rule = Rules(
        id=str(uuid.uuid4()), stage=rule.stage, conditions_op=rule.operation, conditions=conditions, actions=actions
    )
    s.add(database_rule)
    if run_immediately:
        for t in get_transactions(s):
            if rule.run(t):
                s.add(t)
    return database_rule


def get_schedules(s: Session, name: str = None, include_deleted: bool = False) -> typing.Sequence[Schedules]:
    """
    Returns a list of all available schedules.

    :param s: session from Actual local database.
    :param name: pattern name of the payee, case-insensitive.
    :param include_deleted: includes all payees which were deleted via frontend. They would not show normally.
    :return: list of schedules.
    """
    query = base_query(Schedules, name, include_deleted)
    return s.exec(query).all()


def get_or_create_clock(s: Session, client: HULC_Client = None) -> MessagesClock:
    """Loads the HULC Clock from the database. This clock tells the server from when the messages should be
    retrieved. See the [original implementation.](
    https://github.com/actualbudget/actual/blob/5bcfc71be67c6e7b7c8b444e4c4f60da9ea9fdaa/packages/loot-core/src/server/db/index.ts#L81-L98)

    If the clock is not existing, it will be created based on the passed client. If the client is missing, an empty
    client is created. If the clock was already existing, the timestamp will only be overwritten if a client is
    provided, otherwise the original value will be returned.

    :param s: session from Actual local database.
    :param client: HULC Client object.
    :return: The message clock object.
    """
    clock = s.exec(select(MessagesClock)).one_or_none()
    if not clock:
        clock = MessagesClock(id=1)
        if not client:
            client = HULC_Client()  # create a default client
        clock.set_clock({"timestamp": str(client), "merkle": {}})
        s.add(clock)
    else:
        # update the clock only if the client was provided
        if client:
            clock.set_timestamp(client)
    return clock


def get_preferences(s: Session) -> typing.Sequence[Preferences]:
    """
    Loads the preference list from the database.

    :param s: session from Actual local database.
    :return: List of preferences.
    """
    return s.exec(select(Preferences)).all()


def get_or_create_preference(s: Session, key: str, value: str) -> Preferences:
    """
    Loads the preference list from the database. If the key is missing, a new one is created, otherwise it's updated.

    :param s: session from Actual local database.
    :param key: key of the preference.
    :param value: value of the preference.
    :return: the preference object.
    """
    preference = get_preference(s, key)
    if preference is None:
        preference = Preferences(id=key, value=value)
        s.add(preference)
    else:
        preference.value = value
    return preference


def get_preference(s: Session, key: str, default: str = None) -> typing.Optional[Preferences]:
    """
    Gets an existing preference by key name, returns `None` if not found.

    :param s: session from Actual local database.
    :param key: preference name.
    :param default: default value to be returned if key is not found.
    :return: preference matching the key provided. If not found, returns `None`."""
    return s.exec(select(Preferences).where(Preferences.id == key)).one_or_none() or default



from __future__ import annotations

import datetime
import decimal
import enum
import re
import typing
import unicodedata

import pydantic

from actual import ActualError
from actual.crypto import is_uuid
from actual.database import BaseModel, Transactions, get_attribute_by_table_name
from actual.exceptions import ActualSplitTransactionError
from actual.schedules import Schedule


def get_normalized_string(value: str) -> typing.Optional[str]:
    """Normalization of string for comparison. Uses lowercase and Canonical Decomposition.

    See https://github.com/actualbudget/actual/blob/a22160579d6e1f7a17213561cec79c321a14525b/packages/loot-core/src/shared/normalisation.ts
    """
    if value is None:
        return None
    return unicodedata.normalize("NFD", value.lower())


class ConditionType(enum.Enum):
    IS = "is"
    IS_APPROX = "isapprox"
    GT = "gt"
    GTE = "gte"
    LT = "lt"
    LTE = "lte"
    CONTAINS = "contains"
    ONE_OF = "oneOf"
    IS_NOT = "isNot"
    DOES_NOT_CONTAIN = "doesNotContain"
    NOT_ONE_OF = "notOneOf"
    IS_BETWEEN = "isbetween"
    MATCHES = "matches"
    HAS_TAGS = "hasTags"


class ActionType(enum.Enum):
    SET = "set"
    SET_SPLIT_AMOUNT = "set-split-amount"
    LINK_SCHEDULE = "link-schedule"
    PREPEND_NOTES = "prepend-notes"
    APPEND_NOTES = "append-notes"


class BetweenValue(pydantic.BaseModel):
    """Used for `isbetween` rules."""

    num_1: typing.Union[int, float] = pydantic.Field(alias="num1")
    num_2: typing.Union[int, float] = pydantic.Field(alias="num2")

    def __str__(self):
        return f"({self.num_1}, {self.num_2})"

    @pydantic.model_validator(mode="after")
    def convert_value(self):
        if isinstance(self.num_1, float):
            self.num_1 = int(self.num_1 * 100)
        if isinstance(self.num_2, float):
            self.num_2 = int(self.num_2 * 100)
        # sort the values
        self.num_1, self.num_2 = sorted((self.num_1, self.num_2))
        return self


class ValueType(enum.Enum):
    DATE = "date"
    ID = "id"
    STRING = "string"
    NUMBER = "number"
    BOOLEAN = "boolean"
    IMPORTED_PAYEE = "imported_payee"

    def is_valid(self, operation: ConditionType) -> bool:
        """Returns if a conditional operation for a certain type is valid. For example, if the value is of type string,
        then `RuleValueType.STRING.is_valid(ConditionType.GT)` will return false, because there is no logical
        greater than defined for strings."""
        if self == ValueType.DATE:
            return operation.value in ("is", "isapprox", "gt", "gte", "lt", "lte")
        elif self in (ValueType.STRING, ValueType.IMPORTED_PAYEE):
            return operation.value in (
                "is",
                "contains",
                "oneOf",
                "isNot",
                "doesNotContain",
                "notOneOf",
                "matches",
                "hasTags",
            )
        elif self == ValueType.ID:
            return operation.value in ("is", "isNot", "oneOf", "notOneOf")
        elif self == ValueType.NUMBER:
            return operation.value in ("is", "isapprox", "isbetween", "gt", "gte", "lt", "lte")
        else:
            # must be BOOLEAN
            return operation.value in ("is",)

    def validate(self, value: typing.Union[int, typing.List[str], str, None], operation: ConditionType = None) -> bool:
        if isinstance(value, list) and operation in (ConditionType.ONE_OF, ConditionType.NOT_ONE_OF):
            return all(self.validate(v, None) for v in value)
        if value is None:
            return True
        if self == ValueType.ID:
            # make sure it's an uuid
            return isinstance(value, str) and is_uuid(value)
        elif self in (ValueType.STRING, ValueType.IMPORTED_PAYEE):
            return isinstance(value, str)
        elif self == ValueType.DATE:
            try:
                res = bool(get_value(value, self))
            except ValueError:
                res = False
            return res
        elif self == ValueType.NUMBER:
            if operation == ConditionType.IS_BETWEEN:
                return isinstance(value, BetweenValue)
            else:
                return isinstance(value, int)
        else:
            # must be BOOLEAN
            return isinstance(value, bool)

    @classmethod
    def from_field(cls, field: str | None) -> ValueType:
        if field in ("acct", "category", "description"):
            return ValueType.ID
        elif field in ("notes",):
            return ValueType.STRING
        elif field in ("imported_description",):
            return ValueType.IMPORTED_PAYEE
        elif field in ("date",):
            return ValueType.DATE
        elif field in ("cleared", "reconciled"):
            return ValueType.BOOLEAN
        elif field in ("amount",):
            return ValueType.NUMBER
        else:
            raise ValueError(f"Field '{field}' does not have a matching ValueType.")


def get_value(
    value: typing.Union[int, typing.List[str], str, None], value_type: ValueType
) -> typing.Union[int, datetime.date, typing.List[str], str, None]:
    """Converts the value to an actual value according to the type."""
    if value_type is ValueType.DATE:
        if isinstance(value, str):
            return datetime.datetime.strptime(value, "%Y-%m-%d").date()
        elif isinstance(value, int):
            return datetime.datetime.strptime(str(value), "%Y%m%d").date()
    elif value_type is ValueType.BOOLEAN:
        return int(value)  # database accepts 0 or 1
    elif value_type in (ValueType.STRING, ValueType.IMPORTED_PAYEE):
        if isinstance(value, list):
            return [get_value(v, value_type) for v in value]
        else:
            return get_normalized_string(value)
    return value


def condition_evaluation(
    op: ConditionType,
    true_value: typing.Union[int, typing.List[str], str, datetime.date, None],
    self_value: typing.Union[int, typing.List[str], str, datetime.date, BetweenValue, None],
    options: dict = None,
) -> bool:
    """Helper function to evaluate the condition based on the true_value, value found on the transaction, and the
    self_value, value defined on rule condition."""
    if true_value is None:
        # short circuit as comparisons with NoneType are useless
        return False
    if isinstance(options, dict):
        # short circuit if the transaction should be and in/outflow but it isn't
        if options.get("outflow") is True and true_value > 0:
            return False
        if options.get("inflow") is True and true_value < 0:
            return False
    if isinstance(self_value, int) and isinstance(options, dict) and options.get("outflow") is True:
        # if it's an outflow we use the negative value of self_value, that is positive
        self_value = -self_value
    # do comparison
    if op == ConditionType.IS:
        return self_value == true_value
    elif op == ConditionType.IS_NOT:
        return self_value != true_value
    elif op == ConditionType.IS_APPROX:
        if isinstance(true_value, datetime.date):
            # Actual uses two days as reference
            # https://github.com/actualbudget/actual/blob/98a7aac73667241da350169e55edd2fc16a6687f/packages/loot-core/src/server/accounts/rules.ts#L302-L304
            interval = datetime.timedelta(days=2)
            if isinstance(self_value, Schedule):
                return self_value.is_approx(true_value, interval)
        else:
            # Actual uses 7.5% of the value as threshold
            # https://github.com/actualbudget/actual/blob/243703b2f70532ec1acbd3088dda879b5d07a5b3/packages/loot-core/src/shared/rules.ts#L261-L263
            interval = round(abs(self_value) * 0.075, 2)
        return self_value - interval <= true_value <= self_value + interval
    elif op == ConditionType.ONE_OF:
        return true_value in self_value
    elif op == ConditionType.CONTAINS:
        return self_value in true_value
    elif op == ConditionType.MATCHES:
        return bool(re.search(self_value, true_value, re.IGNORECASE))
    elif op == ConditionType.NOT_ONE_OF:
        return true_value not in self_value
    elif op == ConditionType.DOES_NOT_CONTAIN:
        return self_value not in true_value
    elif op == ConditionType.GT:
        return true_value > self_value
    elif op == ConditionType.GTE:
        return true_value >= self_value
    elif op == ConditionType.LT:
        return self_value > true_value
    elif op == ConditionType.LTE:
        return self_value >= true_value
    elif op == ConditionType.IS_BETWEEN:
        return self_value.num_1 <= true_value <= self_value.num_2
    elif op == ConditionType.HAS_TAGS:
        # this regex is not correct, but is good enough according to testing
        # taken from https://stackoverflow.com/a/26740753/12681470
        tags = re.findall(r"\#[\U00002600-\U000027BF\U0001f300-\U0001f64F\U0001f680-\U0001f6FF\w-]+", self_value)
        return any(tag in true_value for tag in tags)
    else:
        raise ActualError(f"Operation {op} not supported")


class Condition(pydantic.BaseModel):
    """
    A condition does a single comparison check for a transaction. The `op` indicates the action type, usually being
    set to `IS` or `CONTAINS`, and the comparison is applied to a `field` with certain `value`. If the transaction
    value matches the condition, the `run` method returns `True`, otherwise it returns `False`. The individual condition
    cannot change the value of the transaction, as only the [Action][actual.rules.Action] can.

    **Important**: Actual shows the amount on frontend as decimal but handles it internally as cents. Make sure that, if
    you provide the `amount` rule manually, you either provide number of cents, as an integers, or a float that get
    automatically converted to cents. As an example, `50` will be interpreted as 50 cents, but `50.0` will be
    interpreted as 50 of the currency (or 5000 cents).

    The 'field' can be one of the following ('type' will be set automatically):

    - `imported_description`: `type` must be `string` and `value` any string
    - `acct`: `type` must be `id` and `value` a valid uuid
    - `category`: `type` must be `id` and `value` a valid uuid
    - `date`: `type` must be `date` and `value` a string in the date format `'2024-04-11'`
    - `description`: `type` must be `id` and `value` a valid uuid (means payee_id)
    - `notes`: `type` must be 'string' and `value` any string
    - `amount`: `type` must be `number` and format in cents
    - `amount_inflow`: `type` must be `number` and format in cents, will set `"options":{"inflow":true}`
    - `amount_outflow`: `type` must be `number` and format in cents, will set `"options":{"outflow":true}`
    """

    field: typing.Literal[
        "imported_description",
        "acct",
        "category",
        "date",
        "description",
        "notes",
        "amount",
        "amount_inflow",
        "amount_outflow",
    ]
    op: ConditionType
    value: typing.Union[
        int,
        float,
        str,
        typing.List[str],
        Schedule,
        typing.List[BaseModel],
        BetweenValue,
        BaseModel,
        datetime.date,
        None,
    ]
    type: typing.Optional[ValueType] = None
    options: typing.Optional[dict] = None

    def __str__(self) -> str:
        v = f"'{self.value}'" if isinstance(self.value, str) or isinstance(self.value, Schedule) else str(self.value)
        return f"'{self.field}' {self.op.value} {v}"

    def as_dict(self):
        """Returns valid dict for database insertion."""
        ret = self.model_dump(mode="json")
        if not self.options:
            ret.pop("options", None)
        return ret

    def get_value(self) -> typing.Union[int, datetime.date, typing.List[str], str, None]:
        return get_value(self.value, self.type)

    @pydantic.model_validator(mode="after")
    def convert_value(self):
        if self.field in ("amount_inflow", "amount_outflow") and self.options is None:
            self.options = {self.field.split("_")[1]: True}
            self.value = abs(self.value)
            self.field = "amount"
        if isinstance(self.value, float):
            # convert silently in the background to a valid number
            self.value = int(self.value * 100)
        return self

    @pydantic.model_validator(mode="after")
    def check_operation_type(self):
        if not self.type:
            self.type = ValueType.from_field(self.field)
        # check if types are fine
        if not self.type.is_valid(self.op):
            raise ValueError(f"Operation {self.op} not supported for type {self.type}")
        # if a pydantic object is provided and id is expected, extract the id
        if isinstance(self.value, BaseModel):
            self.value = str(self.value.id)
        elif isinstance(self.value, list) and len(self.value) and isinstance(self.value[0], pydantic.BaseModel):
            self.value = [v.id if hasattr(v, "id") else v for v in self.value]
        # make sure the data matches the value type
        if not self.type.validate(self.value, self.op):
            raise ValueError(f"Value {self.value} is not valid for type {self.type.name} and operation {self.op.name}")
        return self

    def run(self, transaction: Transactions) -> bool:
        attr = get_attribute_by_table_name(Transactions.__tablename__, self.field)
        true_value = get_value(getattr(transaction, attr), self.type)
        self_value = self.get_value()
        return condition_evaluation(self.op, true_value, self_value, self.options)


class Action(pydantic.BaseModel):
    """
    An Action does a single column change for a transaction. The `op` indicates the
    [ActionType][actual.rules.ActionType], usually being to SET a `field` with certain `value`.

    For the `op` `LINKED_SCHEDULE`, the operation will link the transaction to a certain schedule id that generated it.
    This is often done automatically when you create schedules, as they are also rules internally.

    For the `op` `SET_SPLIT_AMOUNT`, the transaction will be split into multiple different splits depending on the
    method defined by the user, being it on `fixed-amount`, `fixed-percent` or `remainder`. The method will be stored
    under `options` with the format `{"method": "remainder", "splitIndex": 1}`, indicating both the method of splitting
    but also which index that split will take.

    The `field` can be one of the following (`type` will be set automatically based on the database):

    - `category`: `type` must be `id` and `value` a valid uuid
    - `description`: `type` must be `id` `value` a valid uuid (additional `"options":{"splitIndex":0}`)
    - `notes`: `type` must be `string` and `value` any string
    - `cleared`: `type` must be `boolean` and `value` is a literal True/False (additional `"options":{"splitIndex":0}`)
    - `acct`: `type` must be `id` and `value` an uuid
    - `date`: `type` must be `date` and `value` a string in the date format `"2024-04-11"`
    - `amount`: `type` must be `number` and format in cents
    """

    field: typing.Optional[typing.Literal["category", "description", "notes", "cleared", "acct", "date", "amount"]] = (
        None
    )
    op: ActionType = pydantic.Field(ActionType.SET, description="Action type to apply (default changes a column).")
    value: typing.Union[str, bool, int, float, pydantic.BaseModel, None]
    type: typing.Optional[ValueType] = None
    options: typing.Dict[str, typing.Union[str, int]] = None

    def __str__(self) -> str:
        if self.op in (ActionType.SET, ActionType.LINK_SCHEDULE):
            split_info = ""
            if self.options and self.options.get("splitIndex") > 0:
                split_info = f" at Split {self.options.get('splitIndex')}"
            field_str = f" '{self.field}'" if self.field else ""
            return f"{self.op.value}{field_str}{split_info} to '{self.value}'"
        elif self.op == ActionType.SET_SPLIT_AMOUNT:
            method = self.options.get("method") or ""
            split_index = self.options.get("splitIndex") or ""
            return f"allocate a {method} at Split {split_index}: {self.value}"
        elif self.op in (ActionType.APPEND_NOTES, ActionType.PREPEND_NOTES):
            return (
                f"append to notes '{self.value}'"
                if self.op == ActionType.APPEND_NOTES
                else f"prepend to notes '{self.value}'"
            )

    def as_dict(self):
        """Returns valid dict for database insertion."""
        ret = self.model_dump(mode="json")
        if not self.options:
            ret.pop("options", None)
        return ret

    @pydantic.model_validator(mode="after")
    def convert_value(self):
        if isinstance(self.value, float):
            # convert silently in the background to a valid number
            self.value = int(self.value * 100)
        if self.field in ("cleared",) and self.value in (0, 1):
            self.value = bool(self.value)
        return self

    @pydantic.model_validator(mode="after")
    def check_operation_type(self):
        if not self.type:
            if self.field is not None:
                self.type = ValueType.from_field(self.field)
            elif self.op == ActionType.LINK_SCHEDULE:
                self.type = ValueType.ID
            elif self.op == ActionType.SET_SPLIT_AMOUNT:
                self.type = ValueType.NUMBER
        # questionable choice from the developers to set it to ID, I hope they fix it at some point, but we change it
        if self.op in (ActionType.APPEND_NOTES, ActionType.PREPEND_NOTES):
            self.type = ValueType.STRING
        # if a pydantic object is provided and id is expected, extract the id
        if isinstance(self.value, pydantic.BaseModel) and hasattr(self.value, "id"):
            self.value = str(self.value.id)
        # make sure the data matches the value type
        if not self.type.validate(self.value):
            raise ValueError(f"Value {self.value} is not valid for type {self.type.name}")
        return self

    def run(self, transaction: Transactions) -> None:
        """Runs the action on the transaction, regardless of the condition. For the condition based rule, see
        [Rule.run][actual.rules.Rule.run]."""
        if self.op == ActionType.SET:
            attr = get_attribute_by_table_name(Transactions.__tablename__, str(self.field))
            value = get_value(self.value, self.type)
            # if the split index is existing, modify instead the split transaction
            split_index = self.options.get("splitIndex", None) if self.options else None
            if split_index and len(transaction.splits) >= split_index:
                transaction = transaction.splits[split_index - 1]
            # set the value
            if self.type == ValueType.DATE:
                transaction.set_date(value)
            elif attr == "payee_id":
                # this has to be handled separately since, when setting a transfer payee, a transfer transaction needs
                # to be created
                from actual.queries import set_transaction_payee

                # get inner session from object
                session = transaction._sa_instance_state.session  # noqa
                set_transaction_payee(session, transaction, value)
            else:
                setattr(transaction, attr, value)
        elif self.op == ActionType.LINK_SCHEDULE:
            transaction.schedule_id = self.value
        # for the notes rule, check if the rule was already applied since actual does not do that.
        # this should ensure the prefix or suffix is not applied multiple times
        elif self.op == ActionType.APPEND_NOTES:
            notes = transaction.notes or ""
            if not notes.endswith(self.value):
                transaction.notes = f"{notes}{self.value}"
        elif self.op == ActionType.PREPEND_NOTES:
            notes = transaction.notes or ""
            if not notes.startswith(self.value):
                transaction.notes = f"{self.value}{notes}"
        else:
            raise ActualError(f"Operation {self.op} not supported")


class Rule(pydantic.BaseModel):
    """Contains an individual rule, with multiple conditions and multiple actions. Can only be used
    on a single transaction at a time. You can evaluate if the rule would activate without doing any changes on the
    transaction by calling the `evaluate` method.

    Note that the frontend refers to the operation as either 'all' or 'any', but stores them as 'and' and 'or'
    respectively on the database. If you provide the frontend values, they will be converted on the background
    automatically.
    """

    conditions: typing.List[Condition] = pydantic.Field(
        ..., description="List of conditions that need to be met (one or all) in order for the actions to be applied."
    )
    operation: typing.Literal["and", "or"] = pydantic.Field(
        "and", description="Operation to apply for the rule evaluation. If 'all' or 'any' need to be evaluated."
    )
    actions: typing.List[Action] = pydantic.Field(..., description="List of actions to apply to the transaction.")
    stage: typing.Literal["pre", "post", None] = pydantic.Field(
        None, description="Stage in which the rule" "will be evaluated (default None)"
    )

    @pydantic.model_validator(mode="before")
    def correct_operation(cls, value):
        """If the user provides the same `all` or `any` that the frontend provides, we fix it silently to `and` and
        `or` respectively."""
        if value.get("operation") == "all":
            value["operation"] = "and"
        elif value.get("operation") == "any":
            value["operation"] = "or"
        return value

    def __str__(self):
        """Returns a readable string representation of the rule."""
        operation = "all" if self.operation == "and" else "any"
        conditions = f" {self.operation} ".join([str(c) for c in self.conditions])
        actions = ", ".join([str(a) for a in self.actions])
        return f"If {operation} of these conditions match {conditions} then {actions}"

    def set_split_amount(self, transaction: Transactions) -> typing.List[Transactions]:
        """Run the rules from setting split amounts."""
        from actual.queries import create_split  # lazy import to prevert circular issues

        # get actions that split the transaction
        split_amount_actions = [action for action in self.actions if action.op == ActionType.SET_SPLIT_AMOUNT]
        if not split_amount_actions or len(transaction.splits) or transaction.is_child:
            return []  # nothing to create
        # get inner session from object
        session = transaction._sa_instance_state.session  # noqa
        # first, do all entries that have fixed values
        split_by_index: typing.List[Transactions] = [None for _ in range(len(split_amount_actions))]  # noqa
        fixed_split_amount_actions = [a for a in split_amount_actions if a.options["method"] == "fixed-amount"]
        remainder = transaction.amount
        for action in fixed_split_amount_actions:
            remainder -= action.value
            split = create_split(session, transaction, decimal.Decimal(action.value) / 100)
            split_by_index[action.options.get("splitIndex") - 1] = split
        # now do the ones with a percentage amount
        percent_split_amount_actions = [a for a in split_amount_actions if a.options["method"] == "fixed-percent"]
        amount_to_distribute = remainder
        for action in percent_split_amount_actions:
            value = round(amount_to_distribute * action.value / 100, 0)
            remainder -= value
            split = create_split(session, transaction, decimal.Decimal(value) / 100)
            split_by_index[action.options.get("splitIndex") - 1] = split
        # now, divide the remainder equally between the entries
        remainder_split_amount_actions = [a for a in split_amount_actions if a.options["method"] == "remainder"]
        if not len(remainder_split_amount_actions) and remainder:
            # create a virtual split that contains the leftover remainders
            split = create_split(session, transaction, decimal.Decimal(remainder) / 100)
            split_by_index.append(split)
        elif len(remainder_split_amount_actions):
            amount_per_remainder_split = round(remainder / len(remainder_split_amount_actions), 0)
            for action in remainder_split_amount_actions:
                split = create_split(session, transaction, decimal.Decimal(amount_per_remainder_split) / 100)
                remainder -= amount_per_remainder_split
                split_by_index[action.options.get("splitIndex") - 1] = split
            # The last non-fixed split will be adjusted for the remainder
            split_by_index[remainder_split_amount_actions[-1].options.get("splitIndex") - 1].amount += remainder
        # make sure the splits are still valid and the sum equals the parent
        if sum(s.amount for s in split_by_index) != transaction.amount:
            raise ActualSplitTransactionError("Splits do not match amount of parent transaction.")
        transaction.is_parent, transaction.is_child = 1, 0
        # make sure the splits are ordered correctly
        for idx, split in enumerate(split_by_index):
            split.sort_order = -idx - 2
        return split_by_index

    def evaluate(self, transaction: Transactions) -> bool:
        """Evaluates the rule on the transaction, without applying any action."""
        op = any if self.operation == "or" else all
        return op(c.run(transaction) for c in self.conditions)

    def run(self, transaction: Transactions) -> bool:
        """Runs the rule on the transaction, calling evaluate, and if the return is `True` then running each of
        the actions."""
        if condition_met := self.evaluate(transaction):
            splits = self.set_split_amount(transaction)
            if splits:
                transaction.splits = splits
            for action in self.actions:
                if action.op == ActionType.SET_SPLIT_AMOUNT:
                    continue  # handle in the create_splits
                action.run(transaction)
        return condition_met


class RuleSet(pydantic.BaseModel):
    """
    A RuleSet is a collection of [Conditions][actual.rules.Condition] and [Actions][actual.rules.Action] that will
    evaluate for one or more transactions.

    The [Conditions][actual.rules.Condition] are list of logical comparisons that will compare fields from the
    transaction to stored values. If all conditions from a Transaction are met (or any, if the Rule has an `or`
    operation), then the actions will be applied.

    The actions are a list of changes that will be applied to the transaction.

    **Full example ruleset**: "If all of these conditions match 'notes' contains 'foo' then set 'notes' to 'bar'"

    To create that rule set, you can do:

    ```python
    RuleSet(rules=[
        Rule(
            # all of these conditions match
            operation="and",
            # 'notes' contains 'foo'
            conditions=[Condition(field="notes", op=ConditionType.CONTAINS, value="foo")],
            # set 'notes' to 'bar'
            actions=[Action(field="notes", value="bar")],
        )
    ])
    ```
    """

    rules: typing.List[Rule] = pydantic.Field(..., description="List of rules to be evaluated on run.")

    def __str__(self):
        return "\n".join([str(r) for r in self.rules])

    def __iter__(self) -> typing.Iterator[Rule]:
        return self.rules.__iter__()

    def _run(
        self,
        transaction: typing.Union[Transactions, typing.List[Transactions]],
        stage: typing.Literal["pre", "post", None],
    ):
        for rule in [r for r in self.rules if r.stage == stage]:
            if isinstance(transaction, list):
                for t in transaction:
                    rule.run(t)
            else:
                rule.run(transaction)

    def run(
        self,
        transaction: typing.Union[Transactions, typing.Sequence[Transactions]],
        stage: typing.Literal["all", "pre", "post", None] = "all",
    ):
        """
        Runs the rules for each and every transaction on the list. If stage is 'all' (default), all rules are run in
        the order `pre` -> `None` -> `post`. You can provide a value to run only a certain stage of rules.

        If necessary, you can also individually select the rules you want to run by initializing a new ruleset from
        the original one, or select individual rules from, by using the list of rules `RuleSet.rules`.
        """
        if stage == "all":
            self._run(transaction, "pre")
            self._run(transaction, None)
            self._run(transaction, "post")
        else:
            self._run(transaction, stage)  # noqa

    def add(self, rule: Rule):
        """Adds a new rule to the ruleset."""
        self.rules.append(rule)



import datetime
import decimal
import json
from datetime import date, timedelta

import pytest

from actual import Actual, ActualError, reflect_model
from actual.database import Notes, ReflectBudgets, ZeroBudgets
from actual.queries import (
    create_account,
    create_budget,
    create_rule,
    create_splits,
    create_transaction,
    create_transfer,
    get_accounts,
    get_accumulated_budgeted_balance,
    get_budgets,
    get_or_create_category,
    get_or_create_clock,
    get_or_create_payee,
    get_or_create_preference,
    get_preferences,
    get_ruleset,
    get_transactions,
    normalize_payee,
    reconcile_transaction,
    set_transaction_payee,
)
from actual.rules import Action, Condition, ConditionType, Rule


def test_account_relationships(session):
    today = date.today()
    bank = create_account(session, "Bank", 5000)
    create_account(session, "Savings")
    landlord = get_or_create_payee(session, "Landlord")
    rent = get_or_create_category(session, "Rent")
    rent_payment = create_transaction(session, today, "Bank", "Landlord", "Paying rent", "Rent", -1200)
    utilities_payment = create_transaction(session, today, "Bank", "Landlord", "Utilities", "Rent", -50)
    create_transfer(session, today, "Bank", "Savings", 200, "Saving money")
    session.commit()
    assert bank.balance == decimal.Decimal(3550)
    assert landlord.balance == decimal.Decimal(-1250)
    assert rent.balance == decimal.Decimal(-1250)
    assert rent_payment.category == rent
    assert len(bank.transactions) == 4  # includes starting balance and one transfer
    assert len(landlord.transactions) == 2
    assert len(rent.transactions) == 2
    # let's now void the utilities_payment
    utilities_payment.delete()
    session.commit()
    assert bank.balance == decimal.Decimal(3600)
    assert landlord.balance == decimal.Decimal(-1200)
    assert rent.balance == decimal.Decimal(-1200)
    assert len(bank.transactions) == 3
    assert len(landlord.transactions) == 1
    assert len(rent.transactions) == 1
    # delete the payee and category
    rent.delete()
    landlord.delete()
    session.commit()
    assert rent_payment.category is None
    assert rent_payment.payee is None
    # find the deleted transaction again
    deleted_transaction = get_transactions(
        session, today - timedelta(days=1), today + timedelta(days=1), "Util", bank, include_deleted=True
    )
    assert [utilities_payment] == deleted_transaction
    assert get_accounts(session, "Bank") == [bank]


def test_transaction(session):
    today = date.today()
    other = create_account(session, "Other")
    coffee = create_transaction(
        session, date=today, account="Other", payee="Starbucks", notes="coffee", amount=float(-9.95)
    )
    session.commit()
    assert coffee.amount == -995
    assert len(other.transactions) == 1
    assert other.balance == decimal.Decimal("-9.95")


def test_transaction_without_payee(session):
    other = create_account(session, "Other")
    tr = create_transaction(session, date=date.today(), account=other)
    assert tr.payee_id is None


def test_transfer(session):
    bank = create_account(session, "Bank", 200)
    savings = create_account(session, "Savings")
    origin, dst = create_transfer(session, date.today(), "Bank", "Savings", 200, "Saving money")
    assert origin.payee_id == savings.payee.id
    assert dst.payee_id == bank.payee.id
    assert bank.balance == decimal.Decimal(0.0)
    assert savings.balance == decimal.Decimal(200.0)


def test_reconcile_transaction(session):
    today = date.today()
    create_account(session, "Bank")
    rent_payment = create_transaction(
        session, today, "Bank", "Landlord", "Paying rent", "Expenses", -1200, imported_id="unique"
    )
    unrelated = create_transaction(
        session, today - timedelta(days=5), "Bank", "Carshop", "Car maintenance", "Car", -1200
    )
    session.commit()
    assert (
        reconcile_transaction(session, today + timedelta(days=1), "Bank", category="Rent", amount=-1200).id
        == rent_payment.id
    )
    session.commit()
    # check if the property was updated
    assert rent_payment.get_date() == today + timedelta(days=1)
    assert rent_payment.category.name == "Rent"
    # should still be able to match if the payee is defined, as the match is stronger
    assert (
        reconcile_transaction(
            session, today - timedelta(days=5), payee="Landlord", account="Bank", amount=-1200, update_existing=False
        ).id
        == rent_payment.id
    )
    # should not be able to match without payee
    assert reconcile_transaction(session, today - timedelta(days=5), account="Bank", amount=-1200).id == unrelated.id
    # regardless of date, the match by unique id should work
    assert (
        reconcile_transaction(
            session,
            today - timedelta(days=30),
            account="Bank",
            amount=-1200,
            imported_id="unique",
            update_existing=False,
        ).id
        == rent_payment.id
    )
    # but if it's too far, it will be a new transaction
    assert reconcile_transaction(session, today - timedelta(days=30), account="Bank", amount=-1200).id not in (
        rent_payment.id,
        unrelated.id,
    )


def test_create_splits(session):
    bank = create_account(session, "Bank")
    t = create_transaction(session, date.today(), bank, category="Dining", amount=-10.0)
    t_taxes = create_transaction(session, date.today(), bank, category="Taxes", amount=-2.5)
    parent_transaction = create_splits(session, [t, t_taxes], notes="Dining")
    # find all children
    trs = get_transactions(session)
    assert len(trs) == 2
    assert t in trs
    assert t_taxes in trs
    assert all(tr.parent == parent_transaction for tr in trs)
    # find all parents
    parents = get_transactions(session, is_parent=True)
    assert len(parents) == 1
    assert len(parents[0].splits) == 2
    # find all with category
    category = get_transactions(session, category="Dining")
    assert len(category) == 1


def test_create_splits_error(session):
    bank = create_account(session, "Bank")
    wallet = create_account(session, "Wallet")
    t1 = create_transaction(session, date.today(), bank, category="Dining", amount=-10.0)
    t2 = create_transaction(session, date.today(), wallet, category="Taxes", amount=-2.5)
    t3 = create_transaction(session, date.today() - timedelta(days=1), bank, category="Taxes", amount=-2.5)
    with pytest.raises(ActualError, match="must be the same for all transactions in splits"):
        create_splits(session, [t1, t2])
    with pytest.raises(ActualError, match="must be the same for all transactions in splits"):
        create_splits(session, [t1, t3])


def test_create_transaction_without_account_error(session):
    with pytest.raises(ActualError):
        create_transaction(session, date.today(), "foo", "")
    with pytest.raises(ActualError):
        create_transaction(session, date.today(), None, "")


def test_rule_insertion_method(session):
    # create one example transaction
    create_transaction(session, date(2024, 1, 4), create_account(session, "Bank"), "")
    session.commit()
    # create and run rule
    action = Action(field="cleared", value=1)
    assert action.as_dict() == {"field": "cleared", "op": "set", "type": "boolean", "value": True}
    condition = Condition(field="date", op=ConditionType.IS_APPROX, value=date(2024, 1, 2))
    assert condition.as_dict() == {"field": "date", "op": "isapprox", "type": "date", "value": "2024-01-02"}
    # test full rule
    rule = Rule(conditions=[condition], actions=[action], operation="all", stage="pre")
    created_rule = create_rule(session, rule, run_immediately=True)
    assert [condition.as_dict()] == json.loads(created_rule.conditions)
    assert [action.as_dict()] == json.loads(created_rule.actions)
    assert created_rule.conditions_op == "and"
    assert created_rule.stage == "pre"
    trs = get_transactions(session)
    assert trs[0].cleared == 1
    session.flush()
    rs = get_ruleset(session)
    assert len(rs.rules) == 1
    assert str(rs) == "If all of these conditions match 'date' isapprox '2024-01-02' then set 'cleared' to 'True'"


@pytest.mark.parametrize(
    "budget_type,budget_table",
    [("rollover", ZeroBudgets), ("report", ReflectBudgets), ("envelope", ZeroBudgets), ("tracking", ReflectBudgets)],
)
def test_budgets(session, budget_type, budget_table):
    # set the config
    get_or_create_preference(session, "budgetType", budget_type)
    # insert a budget
    category = get_or_create_category(session, "Expenses")
    unrelated_category = get_or_create_category(session, "Unrelated")
    session.commit()
    create_budget(session, date(2024, 10, 7), category, 10.0)
    assert len(get_budgets(session)) == 1
    assert len(get_budgets(session, date(2024, 10, 1))) == 1
    assert len(get_budgets(session, date(2024, 10, 1), category)) == 1
    assert len(get_budgets(session, date(2024, 9, 1))) == 0
    budget = get_budgets(session)[0]
    assert isinstance(budget, budget_table)
    assert budget.get_amount() == 10.0
    assert budget.get_date() == date(2024, 10, 1)
    # get a budget that already exists, but re-set it
    create_budget(session, date(2024, 10, 7), category, 20.0)
    assert budget.get_amount() == 20.0
    assert budget.range == (date(2024, 10, 1), date(2024, 11, 1))
    # insert a transaction in the range and see if they are counted on the balance
    bank = create_account(session, "Bank")
    t1 = create_transaction(session, date(2024, 10, 1), bank, category=category, amount=-10.0)
    t2 = create_transaction(session, date(2024, 10, 15), bank, category=category, amount=-10.0)
    t3 = create_transaction(session, date(2024, 10, 31), bank, category=category, amount=-15.0)
    # should not be counted
    create_transaction(session, date(2024, 10, 1), bank, category=category, amount=-15.0).delete()
    create_transaction(session, date(2024, 11, 1), bank, category=category, amount=-20.0)
    create_transaction(session, date(2024, 10, 15), bank, category=unrelated_category, amount=-20.0)
    assert budget.balance == -35.0
    budget_transactions = get_transactions(session, budget=budget)
    assert len(budget_transactions) == 3
    assert all(t in budget_transactions for t in (t1, t2, t3))
    # test if it fails if category does not exist
    with pytest.raises(ActualError, match="Category is provided but does not exist"):
        get_budgets(session, category="foo")
    # filtering by budget will raise a warning if get_transactions with budget also provides a start-end outside range
    with pytest.warns(match="Provided date filters"):
        get_transactions(session, date(2024, 9, 1), date(2024, 9, 15), budget=budget)


@pytest.mark.parametrize(
    "budget_type,with_reset,with_previous_value,expected_value_previous_month,expected_value_current_month",
    [
        ("envelope", False, False, decimal.Decimal(5), decimal.Decimal(25)),
        ("envelope", False, True, decimal.Decimal(15), decimal.Decimal(35)),
        ("envelope", True, True, decimal.Decimal(-5), decimal.Decimal(20)),
        ("envelope", True, False, decimal.Decimal(-15), decimal.Decimal(20)),
        ("tracking", False, True, decimal.Decimal(-5), decimal.Decimal(20)),
    ],
)
def test_accumulated_budget_amount(
    session, budget_type, with_reset, with_previous_value, expected_value_current_month, expected_value_previous_month
):
    get_or_create_preference(session, "budgetType", budget_type)

    category = get_or_create_category(session, "Expenses")
    bank = create_account(session, "Bank")

    # create three months of budgets
    create_budget(session, date(2025, 1, 1), category, 20.0)
    create_budget(session, date(2025, 2, 1), category, 20.0)
    create_budget(session, date(2025, 3, 1), category, 20.0)
    # should be considered since is an income before the beginning of the budget period
    if with_previous_value:
        create_transaction(session, date(2024, 10, 1), bank, category=category, amount=10.0)
    # other transactions
    create_transaction(session, date(2025, 1, 1), bank, category=category, amount=-10.0)
    create_transaction(session, date(2025, 2, 1), bank, category=category, amount=-10.0)
    create_transaction(session, date(2025, 2, 3), bank, category=category, amount=-15.0)
    # should reset rollover budget
    if with_reset:
        create_transaction(session, date(2025, 2, 4), bank, category=category, amount=-20.0)

    assert get_accumulated_budgeted_balance(session, date(2025, 2, 1), category) == expected_value_previous_month
    assert get_accumulated_budgeted_balance(session, date(2025, 3, 1), category) == expected_value_current_month


def test_normalize_payee():
    assert normalize_payee("   mY paYeE ") == "My Payee"
    assert normalize_payee("  ", raw_payee_name=True) == ""
    assert normalize_payee(" My PayeE ", raw_payee_name=True) == "My PayeE"


def test_rollback(session):
    create_account(session, "Bank", 5000)
    session.flush()
    assert "messages" in session.info
    assert len(session.info["messages"])
    session.rollback()
    assert "messages" not in session.info


def test_model_notes(session):
    account_with_note = create_account(session, "Bank 1")
    account_without_note = create_account(session, "Bank 2")
    session.add(Notes(id=f"account-{account_with_note.id}", note="My note"))
    session.commit()
    assert account_with_note.notes == "My note"
    assert account_without_note.notes is None


def test_default_imported_payee(session):
    t = create_transaction(session, date(2024, 1, 4), create_account(session, "Bank"), imported_payee=" foo ")
    session.flush()
    assert t.payee.name == "foo"
    assert t.imported_description == "foo"


def test_session_error(mocker):
    mocker.patch("actual.Actual.validate")
    with Actual(token="foo") as actual:
        with pytest.raises(ActualError, match="No session defined"):
            print(actual.session)  # try to access the session, should raise an exception


def test_apply_changes(session, mocker):
    mocker.patch("actual.Actual.validate")
    actual = Actual(token="foo")
    actual._session, actual.engine, actual._meta = session, session.bind, reflect_model(session.bind)
    # create elements but do not commit them
    account = create_account(session, "Bank")
    transaction = create_transaction(session, date(2024, 1, 4), account, amount=35.7)
    session.flush()
    messages_size = len(session.info["messages"])
    transaction.notes = "foobar"
    session.flush()
    assert len(session.info["messages"]) == messages_size + 1
    messages = session.info["messages"]
    # undo all changes, but apply via database
    session.rollback()
    actual.apply_changes(messages)
    # make sure elements got committed correctly
    accounts = get_accounts(session, "Bank")
    assert len(accounts) == 1
    assert accounts[0].id == account.id
    assert accounts[0].name == account.name
    transactions = get_transactions(session)
    assert len(transactions) == 1
    assert transactions[0].id == transaction.id
    assert transactions[0].notes == transaction.notes
    assert transactions[0].get_date() == transaction.get_date()
    assert transactions[0].get_amount() == transaction.get_amount()


def test_get_or_create_clock(session):
    clock = get_or_create_clock(session)
    assert clock.get_timestamp().ts == datetime.datetime(1970, 1, 1, 0, 0, 0)
    assert clock.get_timestamp().initial_count == 0


def test_get_preferences(session):
    assert len(get_preferences(session)) == 0
    preference = get_or_create_preference(session, "foo", "bar")
    assert preference.value == "bar"
    preferences = get_preferences(session)
    assert len(preferences) == 1
    assert preferences[0] == preference
    # update preference
    get_or_create_preference(session, "foo", "foobar")
    new_preferences = get_preferences(session)
    assert len(new_preferences) == 1
    assert new_preferences[0].value == "foobar"


def test_set_payee_to_transfer(session):
    wallet = create_account(session, "Wallet")
    bank = create_account(session, "Bank")
    session.commit()
    # Create a transaction setting the payee
    t = create_transaction(session, date.today(), bank, wallet.payee, amount=-50)
    session.commit()
    transactions = get_transactions(session)
    assert len(transactions) == 2
    assert transactions[0].get_amount() == -transactions[1].get_amount()
    assert transactions[0].transferred_id == transactions[1].id
    assert transactions[1].transferred_id == transactions[0].id
    # Set this payee to something else, transaction should be deleted
    set_transaction_payee(session, t, None)
    session.commit()
    assert len(get_transactions(session)) == 1
    assert t.payee_id is None
    assert t.transferred_id is None
    # Set payee_id back, transaction should be recreated
    set_transaction_payee(session, t, wallet.payee.id)
    session.commit()
    assert t.payee_id == wallet.payee.id
    assert t.transfer.transfer == t



import datetime

import pytest
from sqlalchemy import delete, select
from testcontainers.core.container import DockerContainer
from testcontainers.core.waiting_utils import wait_for_logs

from actual import Actual, js_migration_statements
from actual.database import __TABLE_COLUMNS_MAP__, Dashboard, Migrations, reflect_model
from actual.exceptions import ActualDecryptionError, ActualError, AuthorizationError
from actual.queries import (
    create_transaction,
    get_accounts,
    get_categories,
    get_or_create_account,
    get_or_create_category,
    get_or_create_payee,
    get_payees,
    get_rules,
    get_ruleset,
    get_schedules,
    get_transactions,
)

VERSIONS = ["25.4.0", "25.5.0", "25.6.1", "25.7.1"]


@pytest.fixture(params=VERSIONS)  # todo: support multiple versions at once
def actual_server(request):
    # we test integration with the 5 latest versions of actual server
    with DockerContainer(f"actualbudget/actual-server:{request.param}").with_exposed_ports(5006) as container:
        wait_for_logs(container, "Listening on :::5006...")
        yield container


def test_create_user_file(actual_server):
    port = actual_server.get_exposed_port(5006)
    with Actual(f"http://localhost:{port}", password="mypass", bootstrap=True) as actual:
        assert len(actual.list_user_files().data) == 0
        actual.create_budget("My Budget")
        actual.upload_budget()
        assert "userId" in actual.get_metadata()
        # add some entries to the budget
        acct = get_or_create_account(actual.session, "Bank")
        assert acct.balance == 0
        payee = get_or_create_payee(actual.session, "Landlord")
        category = get_or_create_category(actual.session, "Rent", "Fixed Costs")
        create_transaction(actual.session, datetime.date(2024, 5, 22), acct, payee, "Paying rent", category, -500)
        actual.commit()
        assert acct.balance == -500
        # list user files
        new_user_files = actual.list_user_files().data
        assert len(new_user_files) == 1
        assert new_user_files[-1].name == "My Budget"
        assert actual.info().build is not None
        # run rules
        actual.run_rules()
        # run bank sync
        assert actual.run_bank_sync() == []
        # check also bank sync accounts, should fail because of no token
        assert actual.bank_sync_accounts("simplefin").data.error_type == "INVALID_ACCESS_TOKEN"
        # same test with goCardless returns 404 for some reason, so we don't do that

    # make sure a new instance can now retrieve the budget info
    with Actual(f"http://localhost:{port}", password="mypass", file="My Budget"):
        assert len(get_accounts(actual.session)) == 1
        assert len(get_payees(actual.session)) == 2  # one is the account payee
        assert len(get_categories(actual.session)) > 6  # there are 6 default categories
        assert len(get_transactions(actual.session)) == 1
        assert len(get_schedules(actual.session)) == 0
        assert len(get_rules(actual.session)) == 0
        assert get_ruleset(actual.session).rules == []

    with pytest.raises(AuthorizationError):
        Actual(actual.api_url, password="mywrongpass", file="My Budget")


def test_encrypted_file(actual_server):
    port = actual_server.get_exposed_port(5006)
    with Actual(f"http://localhost:{port}", password="mypass", encryption_password="mypass", bootstrap=True) as actual:
        actual.create_budget("My Encrypted Budget")
        actual.upload_budget()
        user_files = actual.list_user_files().data
        assert user_files[0].encrypt_key_id is not None
    # re-download budget
    with Actual(
        f"http://localhost:{port}", password="mypass", encryption_password="mypass", file="My Encrypted Budget"
    ) as actual:
        assert actual.session is not None
    with pytest.raises(ActualDecryptionError, match="Error decrypting file. Is the encryption key correct"):
        Actual(
            f"http://localhost:{port}", password="mypass", encryption_password="mywrongpass", file="My Encrypted Budget"
        ).download_budget()
    with pytest.raises(ActualDecryptionError, match="File is encrypted but no encryption password was provided"):
        Actual(f"http://localhost:{port}", password="mypass", file="My Encrypted Budget").download_budget()


def test_update_file_name(actual_server):
    port = actual_server.get_exposed_port(5006)
    with Actual(f"http://localhost:{port}", password="mypass", bootstrap=True) as actual:
        assert len(actual.list_user_files().data) == 0
        actual.create_budget("My Budget")
        actual.upload_budget()
        actual.rename_budget("Other name")
        files = actual.list_user_files().data
        assert len(files) == 1
        assert files[0].name == "Other name"
    # should raise an error if budget does not exist
    with Actual(f"http://localhost:{port}", password="mypass") as actual:
        with pytest.raises(ActualError):
            actual.rename_budget("Failing name")


def test_reimport_file_from_zip(actual_server, tmp_path):
    port = actual_server.get_exposed_port(5006)
    backup_file = f"{tmp_path}/backup.zip"
    # create one file
    with Actual(f"http://localhost:{port}", password="mypass", bootstrap=True) as actual:
        # add some entries to the budget
        actual.create_budget("My Budget")
        get_or_create_account(actual.session, "Bank")
        actual.commit()
        actual.upload_budget()
    # re-download file and save as a backup
    with Actual(f"http://localhost:{port}", password="mypass", file="My Budget") as actual:
        actual.export_data(backup_file)
        actual.delete_budget()
    # re-upload the file
    with Actual(f"http://localhost:{port}", password="mypass") as actual:
        actual.import_zip(backup_file)
        actual.upload_budget()
    # check if the account can be retrieved
    with Actual(f"http://localhost:{port}", password="mypass", file="My Budget") as actual:
        assert len(get_accounts(actual.session)) == 1


def test_redownload_file(actual_server, tmp_path):
    port = actual_server.get_exposed_port(5006)
    with Actual(f"http://localhost:{port}", password="mypass", bootstrap=True) as actual:
        actual.create_budget("My Budget")
        actual.upload_budget()
        file_id = actual.get_metadata()["cloudFileId"]
    # do a normal download and see if the folder matches the fileId that was initially generated
    with Actual(f"http://localhost:{port}", password="mypass", file="My Budget") as actual:
        assert str(actual._data_dir).endswith(file_id)
    # download to a certain folder
    with Actual(f"http://localhost:{port}", password="mypass", file="My Budget", data_dir=tmp_path) as actual:
        get_or_create_account(actual.session, "Bank")
        actual.commit()
        assert not str(actual._data_dir).endswith(file_id)
    # reupload the budget
    with Actual(f"http://localhost:{port}", password="mypass", file="My Budget", data_dir=tmp_path) as actual:
        actual.reupload_budget()
    with pytest.warns(match="Sync id has been reset on remote database, re-downloading the budget"):
        with Actual(f"http://localhost:{port}", password="mypass", file="My Budget", data_dir=tmp_path):
            pass


def test_reset_password(actual_server):
    port = actual_server.get_exposed_port(5006)
    with Actual(f"http://localhost:{port}", password="mypass", bootstrap=True) as actual:
        actual.create_budget("My Budget")
        actual.upload_budget()
        actual.reset_password("mynewpass")
        response = actual.list_user_files()
        assert len(response.data) == 1
    with Actual(f"http://localhost:{port}", password="mynewpass"):
        assert len(actual.list_user_files().data) == 1
        with pytest.raises(AuthorizationError):
            # login with old password should fail
            actual.login("mypass")


def test_models(actual_server):
    port = actual_server.get_exposed_port(5006)
    with Actual(f"http://localhost:{port}", password="mypass", encryption_password="mypass", bootstrap=True) as actual:
        actual.create_budget("My Budget")
        # check if the models are matching
        metadata = reflect_model(actual.session.bind)
        # check first if all tables are present
        for table_name, table in metadata.tables.items():
            assert table_name in __TABLE_COLUMNS_MAP__, f"Missing table '{table_name}' on models."
            # then assert if all columns are matching the model
            for column_name in table.columns.keys():
                assert (
                    column_name in __TABLE_COLUMNS_MAP__[table_name]["columns"]
                ), f"Missing column '{column_name}' at table '{table_name}'."


def test_header_login():
    # TODO: this is fixed on a previous version since header login doesn't seem to be working fully on latest version
    working_version = "25.3.0"
    with (
        DockerContainer(f"actualbudget/actual-server:{working_version}")
        .with_env("ACTUAL_LOGIN_METHOD", "header")
        .with_exposed_ports(5006) as container
    ):
        port = container.get_exposed_port(5006)
        wait_for_logs(container, "Listening on :::5006...")
        with Actual(f"http://localhost:{port}", password="mypass", bootstrap=True):
            pass
        # make sure we can log in
        actual = Actual(f"http://localhost:{port}", password="mypass")
        response_login = actual.login("mypass")
        response_header_login = actual.login("mypass", "header")
        assert response_login.data.token == response_header_login.data.token


def test_session_reflection_after_migrations():
    with DockerContainer(f"actualbudget/actual-server:{VERSIONS[-1]}").with_exposed_ports(5006) as container:
        port = container.get_exposed_port(5006)
        wait_for_logs(container, "Listening on :::5006...")
        with Actual(f"http://localhost:{port}", password="mypass", bootstrap=True) as actual:
            actual.create_budget("My Budget")
            actual.upload_budget()
            # add a dashboard entry
            actual.session.add(Dashboard(id="123", x=1, y=2))
            actual.commit()
            # revert the dashboard creation migration like it never happened
            Dashboard.__table__.drop(actual.engine)
            actual.session.exec(delete(Migrations).where(Migrations.id == 1722804019000))
            actual.session.commit()
        # now try to download the budget, it should not fail
        with Actual(f"http://localhost:{port}", file="My Budget", password="mypass") as actual:
            assert len(actual.session.exec(select(Dashboard)).all()) > 2  # there are two default dashboards


def test_empty_query_migrations():
    # empty queries should not fail
    assert js_migration_statements("await db.runQuery('');") == []
    # malformed entries should not fail
    assert js_migration_statements("await db.runQuery(") == []
    # weird formats neither
    assert js_migration_statements("db.runQuery\n('update 1')") == ["update 1;"]



import pathlib
import tempfile
from typing import Union


def get_base_tmp_folder() -> pathlib.Path:
    """Returns the temporary folder that the library should use to store temporary files if the user does not provide
    a folder path."""
    base_tmp_dir = pathlib.Path(tempfile.gettempdir())
    tmp_dir = base_tmp_dir / "actualpy"
    tmp_dir.mkdir(exist_ok=True)
    return tmp_dir


def get_tmp_folder(file_id: Union[str, None]) -> pathlib.Path:
    """Returns a base folder to store the file based on the file id. Will create the folder if not existing."""
    if not file_id:
        folder = pathlib.Path(tempfile.mkdtemp())
    else:
        folder = get_base_tmp_folder() / str(file_id)
        folder.mkdir(exist_ok=True)
    return folder



from __future__ import annotations

import base64
import datetime
import io
import json
import pathlib
import sqlite3
import uuid
import warnings
import zipfile
from os import PathLike
from typing import IO, List, Optional, Union

from sqlmodel import MetaData, Session, create_engine

from actual.api import ActualServer
from actual.api.models import BankSyncErrorDTO, RemoteFileListDTO
from actual.crypto import create_key_buffer, decrypt_from_meta, encrypt, make_salt
from actual.database import (
    Accounts,
    Transactions,
    apply_change,
    get_attribute_from_reflected_table_name,
    get_class_from_reflected_table_name,
    reflect_model,
    strong_reference_session,
)
from actual.exceptions import (
    ActualBankSyncError,
    ActualDecryptionError,
    ActualError,
    InvalidZipFile,
    UnknownFileId,
)
from actual.migrations import js_migration_statements
from actual.protobuf_models import HULC_Client, Message, SyncRequest
from actual.queries import (
    create_transaction,
    get_account,
    get_accounts,
    get_or_create_clock,
    get_or_create_payee,
    get_ruleset,
    get_transactions,
    reconcile_transaction,
)
from actual.utils.storage import get_tmp_folder
from actual.version import __version__  # noqa: F401


class Actual(ActualServer):
    def __init__(
        self,
        base_url: str = "http://localhost:5006",
        token: str = None,
        password: str = None,
        file: str = None,
        encryption_password: str = None,
        data_dir: Union[str, pathlib.Path] = None,
        cert: str | bool = None,
        bootstrap: bool = False,
        sa_kwargs: dict = None,
        extra_headers: dict[str, str] = None,
    ):
        """
        Implements the Python API for the Actual Server in order to be able to read and modify information on Actual
        books using Python.

        Parts of the implementation are [available at the following file.](
        https://github.com/actualbudget/actual/blob/2178da0414958064337b2c53efc95ff1d3abf98a/packages/loot-core/src/server/cloud-storage.ts)

        :param base_url: url of the running Actual server
        :param token: the token for authentication, if this is available (optional)
        :param password: the password for authentication. It will be used on the .login() method to retrieve the token.
        :param file: the name or id of the file to be set
        :param encryption_password: password used to configure encryption, if existing
        :param data_dir: where to store the downloaded files from the server. If not specified, a temporary folder will
                         be created instead. If database files are already present on the path, the library will try to
                         reuse them by re-computing the sync request.
        :param cert: if a custom certificate should be used (i.e. self-signed certificate), it's path can be provided
                     as a string. Set to `False` for no certificate check.
        :param bootstrap: if the server is not bootstrapped, bootstrap it with the password.
        :param sa_kwargs: additional kwargs passed to the SQLAlchemy session maker. Examples are `autoflush` (enabled
        by default), `autocommit` (disabled by default). For a list of all parameters, check the [SQLAlchemy
        documentation.](https://docs.sqlalchemy.org/en/20/orm/session_api.html#sqlalchemy.orm.Session.__init__)
        :param extra_headers: additional headers to be attached to each request to the Actual server
        """
        super().__init__(base_url, token, password, bootstrap, cert, extra_headers)
        self._file: RemoteFileListDTO | None = None
        self._data_dir = pathlib.Path(data_dir) if data_dir else None
        self.engine = None
        self._session: Session | None = None
        self._client: HULC_Client | None = None
        self._meta: MetaData | None = None  # stores the metadata loaded from remote
        # set the correct file
        if file:
            self.set_file(file)
        self._encryption_password = encryption_password
        self._master_key = None
        self._in_context = False
        self._sa_kwargs = sa_kwargs or {}
        if "autoflush" not in self._sa_kwargs:
            self._sa_kwargs["autoflush"] = True

    def __enter__(self) -> Actual:
        self._in_context = True
        if self._file:
            self.download_budget(self._encryption_password)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._session:
            self._session.close()
        self._in_context = False

    @property
    def session(self) -> Session:
        if not self._session:
            raise ActualError(
                "No session defined. Use `with Actual() as actual:` construct to generate one.\n"
                "If you are already using the context manager, try setting a file to use the session."
            )
        return self._session

    def set_file(self, file_id: Union[str, RemoteFileListDTO]) -> RemoteFileListDTO:
        """
        Sets the file id for the class for further requests. The file_id argument can be either the name, the remote
        id or the group id (also known as sync_id) from the file. If there are duplicates for the name, this method
        will raise `UnknownFileId`.
        """
        if isinstance(file_id, RemoteFileListDTO):
            self._file = file_id
            return file_id
        selected_files = []
        user_files = self.list_user_files()
        for file in user_files.data:
            if (file.file_id == file_id or file.name == file_id or file.group_id == file_id) and file.deleted == 0:
                selected_files.append(file)
        if len(selected_files) == 0:
            raise UnknownFileId(f"Could not find a file id or identifier '{file_id}'")
        elif len(selected_files) > 1:
            raise UnknownFileId(f"Multiple files found with identifier '{file_id}'")
        return self.set_file(selected_files[0])

    def run_migrations(self, migration_files: List[str]):
        """Runs the migration files, skipping the ones that have already been run. The files can be retrieved from
        .data_file_index() method. This first file is the base database, and the following files are migrations.
        Migrations can also be .js files. In this case, we have to extract and execute queries from the standard JS."""
        conn = sqlite3.connect(self._data_dir / "db.sqlite")
        for file in migration_files:
            if not file.startswith("migrations"):
                continue  # in case db.sqlite file gets passed as one of the migrations files
            file_id = file.split("_")[0].split("/")[1]
            if conn.execute(f"SELECT id FROM __migrations__ WHERE id = '{file_id}';").fetchall():
                continue  # skip migration as it was already ran
            migration = self.data_file(file)  # retrieves file from actual server
            sql_statements = migration.decode()
            if file.endswith(".js"):
                # there is one migration which is Javascript. All entries inside db.execQuery(`...`) must be executed
                exec_entries = js_migration_statements(sql_statements)
                sql_statements = "\n".join(exec_entries)
            conn.executescript(sql_statements)
            conn.execute(f"INSERT INTO __migrations__ (id) VALUES ({file_id});")
        conn.commit()
        conn.close()
        # update the metadata by reflecting the model
        self._meta = reflect_model(self.engine)

    def create_budget(self, budget_name: str):
        """Creates a budget using the remote server default database and migrations. If password is provided, the
        budget will be encrypted. It's important to note that `create_budget` depends on the migration files from the
        Actual server, and those could be written in Javascript. Event though the library tries to execute all
        statements in those files, is not an exact match. It is preferred to create budgets via frontend instead."""
        warnings.warn("Creating budgets via actualpy is not recommended due to custom code migrations.")
        migration_files = self.data_file_index()
        file_id = str(uuid.uuid4())
        # create folder for the files
        if not self._data_dir:
            self._data_dir = get_tmp_folder(file_id)
        # first migration file is the default database
        migration = self.data_file(migration_files[0])
        (self._data_dir / "db.sqlite").write_bytes(migration)
        # also write the metadata file with default fields
        random_id = str(uuid.uuid4()).replace("-", "")[:7]
        self.update_metadata(
            {
                "id": f"My-Finances-{random_id}",
                "budgetName": budget_name,
                "userId": self._token,
                "cloudFileId": file_id,
                "resetClock": True,
            }
        )
        self._file = RemoteFileListDTO(name=budget_name, fileId=file_id, groupId=None, deleted=0, encryptKeyId=None)
        # generate a session
        self.engine = create_engine(f"sqlite:///{self._data_dir}/db.sqlite")
        # create engine for downloaded database and run migrations
        self.run_migrations(migration_files[1:])
        if self._in_context:
            self._session = strong_reference_session(Session(self.engine, **self._sa_kwargs))
        # create a clock. Since the clock entry is not tracked, we use a separate session
        with Session(self.engine) as session:
            self._client = HULC_Client()
            get_or_create_clock(session, self._client)
            session.commit()

    def rename_budget(self, budget_name: str):
        """Renames the budget with the given name."""
        if not self._file:
            raise UnknownFileId("No current file loaded.")
        self.update_user_file_name(self._file.file_id, budget_name)

    def delete_budget(self):
        """Deletes the currently loaded file from the server."""
        if not self._file:
            raise UnknownFileId("No current file loaded.")
        self.delete_user_file(self._file.file_id)
        # reset group id, as file cannot be synced anymore
        self._file.group_id = None

    def cleanup(self):
        """
        Cleans up the database from all deleted transactions, message caches and runs a VACUUM. Should reduce the size
        of the database before exporting it.

        Taken from source code at
        [actual/packages/loot-core/src/server/sync/reset.ts](https://github.com/actualbudget/actual/blob/89006275a092d2309ab03162a047e07663789198/packages/loot-core/src/server/sync/reset.ts#L37-L47)
        """
        with sqlite3.connect(self._data_dir / "db.sqlite") as conn:
            conn.executescript(
                """
                DELETE FROM messages_crdt;
                DELETE FROM messages_clock;
                DELETE FROM transactions WHERE tombstone = 1;
                DELETE FROM accounts WHERE tombstone = 1;
                DELETE FROM payees WHERE tombstone = 1;
                DELETE FROM categories WHERE tombstone = 1;
                DELETE FROM category_groups WHERE tombstone = 1;
                DELETE FROM schedules WHERE tombstone = 1;
                DELETE FROM rules WHERE tombstone = 1;
                ANALYZE;
                VACUUM;
            """
            )

    def export_data(self, output_file: str | PathLike[str] | IO[bytes] = None, cleanup: bool = True) -> bytes:
        """Export your data as a zip file containing db.sqlite and metadata.json files. It can be imported into another
        Actual instance by closing an open file (if any), then clicking the “Import file” button, then choosing
        “Actual.” Even when encryption is enabled, the exported zip file will not have any encryption."""
        if cleanup:
            self.cleanup()
        temp_file = io.BytesIO()
        with zipfile.ZipFile(temp_file, "a", zipfile.ZIP_DEFLATED, False) as z:
            z.write(self._data_dir / "db.sqlite", "db.sqlite")
            z.write(self._data_dir / "metadata.json", "metadata.json")
        content = temp_file.getvalue()
        if output_file:
            with open(output_file, "wb") as f:
                f.write(content)
        return content

    def encrypt(self, encryption_password: str):
        """Encrypts the local database using a new key, and re-uploads to the server.

        WARNING: this resets the file on the server. Make sure you have a copy of the database before attempting this
        operation.
        """
        if encryption_password and not self._file.encrypt_key_id:
            # password was provided, but encryption key not, create one
            key_id = str(uuid.uuid4())
            salt = make_salt()
            self.user_create_key(self._file.file_id, key_id, encryption_password, salt)
            self.update_metadata({"encryptKeyId": key_id})
            self._file.encrypt_key_id = key_id
        elif self._file.encrypt_key_id:
            key_info = self.user_get_key(self._file.file_id)
            salt = key_info.data.salt
        else:
            raise ActualError("Budget is encrypted but password was not provided")
        self._master_key = create_key_buffer(encryption_password, salt)
        # encrypt binary data with
        encrypted = encrypt(self._file.encrypt_key_id, self._master_key, self.export_data())
        binary_data = io.BytesIO(base64.b64decode(encrypted["value"]))
        encryption_meta = encrypted["meta"]
        self.reset_user_file(self._file.file_id)
        self.upload_user_file(binary_data.getvalue(), self._file.file_id, self._file.name, encryption_meta)
        self.set_file(self._file.file_id)

    def upload_budget(self):
        """Uploads the current file to the Actual server. If attempting to upload your first budget, make sure you use
        [Actual.create_budget][actual.Actual.create_budget] first.
        """
        if not self._data_dir:
            raise UnknownFileId("No current file loaded.")
        if not self._file:
            file_id = str(uuid.uuid4())
            metadata = self.get_metadata()
            budget_name = metadata.get("budgetName", "My Finances")
            self._file = RemoteFileListDTO(name=budget_name, fileId=file_id, groupId=None, deleted=0, encryptKeyId=None)
        binary_data = io.BytesIO()
        with zipfile.ZipFile(binary_data, "a", zipfile.ZIP_DEFLATED, False) as z:
            z.write(self._data_dir / "db.sqlite", "db.sqlite")
            z.write(self._data_dir / "metadata.json", "metadata.json")
        # we have to first upload the user file so the reference id can be used to generate a new encryption key
        self.upload_user_file(binary_data.getvalue(), self._file.file_id, self._file.name)
        # reset local file id to retrieve the grouping id
        self.set_file(self._file.file_id)
        # encrypt the file and re-upload
        if self._encryption_password or self._master_key or self._file.encrypt_key_id:
            self.encrypt(self._encryption_password)

    def reupload_budget(self):
        """Similar to the reset sync option from the frontend, resets the user file on the backend and re-uploads the
        current copy instead. **This operation can be destructive**, so make sure you generate a copy before
        attempting to re-upload your budget."""
        self.reset_user_file(self._file.file_id)
        self.update_metadata({"groupId": None})  # since we don't know what the new group id will be
        self.upload_budget()

    def apply_changes(self, messages: List[Message]):
        """Applies a list of sync changes, based on what the sync method returned on the remote."""
        if not self.engine:
            raise UnknownFileId("No valid file available, download one with download_budget()")
        with Session(self.engine) as s:
            # use the current value to group updates together to the same row
            current_table, current_id, current_value = None, None, {}
            for message in messages:
                if message.dataset == "prefs":
                    # write it to metadata.json instead
                    self.update_metadata({message.row: message.get_value()})
                    continue
                table = get_class_from_reflected_table_name(self._meta, message.dataset)
                if table is None:
                    raise ActualError(
                        f"Actual found a table not supported by the library: table '{message.dataset}' not found\n"
                    )
                column = get_attribute_from_reflected_table_name(self._meta, message.dataset, message.column)
                if column is None:
                    raise ActualError(
                        f"Actual found a column not supported by the library: "
                        f"column '{message.column}' at table '{message.dataset}' not found\n"
                    )
                # if the current id exists, and it's different from the next one, we update the values
                next_id = message.row
                if current_id and (current_id != next_id or table != current_table):
                    apply_change(s, current_table, current_id, current_value)
                    current_table, current_id, current_value = table, next_id, {column: message.get_value()}
                # otherwise update the cache with the current value
                else:
                    current_table, current_id, current_value[column] = table, next_id, message.get_value()
            # if after finishing all values there is a value left, update it too
            if current_table is not None and current_id is not None and current_value is not None:
                apply_change(s, current_table, current_id, current_value)
            s.commit()

    def get_metadata(self) -> dict:
        """Gets the content of metadata.json."""
        metadata_file = self._data_dir / "metadata.json"
        return json.loads(metadata_file.read_text())

    def update_metadata(self, patch: dict):
        """Updates the metadata.json from the Actual file with the patch fields. The patch is a dictionary that will
        then be merged on the metadata and written again to a file."""
        metadata_file = self._data_dir / "metadata.json"
        if metadata_file.is_file():
            config = self.get_metadata()
            config.update(patch)
        else:
            config = patch
        metadata_file.write_text(json.dumps(config, separators=(",", ":")))

    def download_budget(self, encryption_password: str = None):
        """Downloads the budget file from the remote. After the file is downloaded, the sync endpoint is queries
        for the list of pending changes. The changes are individual row updates, that are then applied on by one to
        the downloaded database state.

        If the budget is password protected, the password needs to be present to download the budget, otherwise it will
        fail.

        When a `data_dir` was provided, the method will try to use the local downloaded copy by first checking if the
        sync id (named group id) remains the same. If it does, then the sync is executed using the stored files.
        Otherwise, the file is re-downloaded.
        """
        # check if file has an encryption key and retrieve it
        encryption_password = encryption_password or self._encryption_password
        self.download_master_encryption_key(encryption_password)
        # then download user file if the data_dir is set and both files are present
        if self._data_dir and all((self._data_dir / path).is_file() for path in ["db.sqlite", "metadata.json"]):
            group_id = self.get_metadata().get("groupId")
            # handle the case where a new group id exists and the file needs to be re-downloaded
            if self._file.group_id != group_id:
                warnings.warn("Sync id has been reset on remote database, re-downloading the budget.")
                (self._data_dir / "db.sqlite").unlink()
                (self._data_dir / "metadata.json").unlink()
                return self.download_budget(encryption_password)
            # resume budget
            self.create_engine()
        else:
            file_bytes = self.download_user_file(self._file.file_id)
            if encryption_password is not None and self._file.encrypt_key_id:
                file_info = self.get_user_file_info(self._file.file_id)
                # decrypt file bytes
                file_bytes = decrypt_from_meta(self._master_key, file_bytes, file_info.data.encrypt_meta)
            self.import_zip(io.BytesIO(file_bytes))
            # sometimes downloaded budgets will not have the groupId
            self.update_metadata({"groupId": self._file.group_id})
        # actual js always calls validation
        self.validate()
        # run migrations if needed
        migration_files = self.data_file_index()
        self.run_migrations(migration_files[1:])
        self.sync()
        # create session if not existing
        if self._in_context and not self._session:
            self._session = strong_reference_session(Session(self.engine, **self._sa_kwargs))

    def download_master_encryption_key(self, encryption_password: str) -> Optional[bytes]:
        """Downloads and assembles the key for decrypting the budget based on the provided encryption password.
        If the user file is not encryption, no key will be returned. If the file was encrypted, the key is assembled
        using the key salt and the password with the PBKDF2HMAC algorithm."""
        if self._file.encrypt_key_id and encryption_password is None:
            raise ActualDecryptionError("File is encrypted but no encryption password was provided.")
        if encryption_password is not None and self._file.encrypt_key_id:
            key_info = self.user_get_key(self._file.file_id)
            self._master_key = create_key_buffer(encryption_password, key_info.data.salt)
        return self._master_key

    def import_zip(self, file_bytes: str | PathLike[str] | IO[bytes]):
        """Imports a zip file as the current database, as well as generating the local reflected session. Enables you
        to inspect backups by loading them directly, instead of unzipping the contents."""
        try:
            zip_file = zipfile.ZipFile(file_bytes)
        except zipfile.BadZipfile as e:
            raise InvalidZipFile(f"Invalid zip file: {e}") from None
        # try to extract the file_id from the metadata.json
        if not self._data_dir:
            try:
                metadata = zip_file.read("metadata.json")
                file_id = json.loads(metadata).get("cloudFileId", None)
            except (KeyError, ValueError):
                file_id = None  # can happen if zip does not contain the file or file is not proper JSON
            self._data_dir = get_tmp_folder(file_id)
        # this should extract 'db.sqlite' and 'metadata.json' to the folder
        zip_file.extractall(self._data_dir)
        self.create_engine()

    def create_engine(self):
        self.engine = create_engine(f"sqlite:///{self._data_dir}/db.sqlite")
        self._meta = reflect_model(self.engine)
        # load the client id
        with Session(self.engine) as session:
            clock = get_or_create_clock(session)
            self._client = clock.get_timestamp()

    def sync(self):
        """Does a sync request and applies all changes that are stored on the server on the local copy of the database.
        Since all changes are retrieved, this function cannot be used for partial changes (since the budget is online).
        """
        # after downloading the budget, some pending transactions still need to be retrieved using sync
        request = SyncRequest(
            {
                "messages": [],
                "fileId": self._file.file_id,
                "groupId": self._file.group_id,
                "keyId": self._file.encrypt_key_id,
            }
        )
        request.set_timestamp(client_id=self._client.client_id, now=self._client.ts)
        changes = self.sync_sync(request)
        self.apply_changes(changes.get_messages(self._master_key))
        # after receiving changes, update the client clock with the latest value
        if changes.messages:
            self._client = HULC_Client.from_timestamp(changes.messages[-1].timestamp)
            # store timestamp also inside database. Session might not be available here, so we create one
            with Session(self.engine) as session:
                get_or_create_clock(session, self._client)

    def commit(self):
        """Adds all pending entries to the local database, and sends a sync request to the remote server to synchronize
        the local changes. It's important to note that this process is not atomic, so if the process is interrupted
        before it completes successfully, the files would end up in a unknown state, leading you to have to redo
        the budget download."""
        if not self._session:
            raise ActualError("No session has been created for the file.")
        # create sync request based on the session reference that is tracked
        req = SyncRequest({"fileId": self._file.file_id, "groupId": self._file.group_id})
        if self._file.encrypt_key_id:
            req.keyId = self._file.encrypt_key_id
        req.set_null_timestamp(client_id=self._client.client_id)
        # flush to database, so that all data is evaluated on the database for consistency
        self._session.flush()
        # first we add all new entries and modify is required
        if "messages" in self._session.info:
            req.set_messages(self._session.info["messages"], self._client, master_key=self._master_key)
        # commit to local database to clear the current flush cache
        self._session.commit()
        # sync all changes to the server
        if self._file.group_id:  # only files with a group id can be synced
            self.sync_sync(req)

    def run_rules(self, transactions: Optional[List[Transactions]] = None):
        """Runs all the stored rules on the database on all transactions, without any filters."""
        if transactions is None:
            transactions = get_transactions(self.session, is_parent=True)
        ruleset = get_ruleset(self.session)
        ruleset.run(transactions)

    def _run_bank_sync_account(
        self, acct: Accounts, start_date: datetime.date, is_first_sync: bool
    ) -> List[Transactions]:
        sync_method = acct.account_sync_source
        account_id = acct.account_id
        requisition_id = acct.bank.bank_id if sync_method == "goCardless" else None
        new_transactions_data = self.bank_sync_transactions(
            sync_method.lower(), account_id, start_date, requisition_id=requisition_id
        )
        if isinstance(new_transactions_data, BankSyncErrorDTO):
            raise ActualBankSyncError(
                new_transactions_data.data.error_type,
                new_transactions_data.data.status,
                new_transactions_data.data.reason,
            )
        new_transactions = new_transactions_data.data.transactions.all
        imported_transactions = []
        if is_first_sync:
            # actual uses 'startingBalance', that already comes in cents and should be enough for our purposes
            # https://github.com/actualbudget/actual/blob/f09f4af667ddd57e031dcdb0d428ae935aa2afad/packages/loot-core/src/server/accounts/sync.ts#L740-L752
            balance_to_use = new_transactions_data.data.balance
            # For simpleFin, the startingBalance is actually the current balance, so we have to use it to deduce the
            # actual startingBalance
            if acct.account_sync_source and acct.account_sync_source.lower() == "simplefin":
                current_balance = new_transactions_data.data.balance
                balance_to_use = current_balance - sum(t.transaction_amount.amount for t in new_transactions)
            if balance_to_use:
                payee = None if acct.offbudget else get_or_create_payee(self.session, "Starting Balance")
                # get date from the oldest transaction. There seems to be a bug here, and it gets the youngest transaction.
                oldest_date = new_transactions[-1].date if new_transactions else datetime.date.today()
                reconciled_transaction = create_transaction(
                    self.session, oldest_date, acct, payee, notes=None, amount=balance_to_use, cleared=True
                )
                reconciled_transaction.starting_balance_flag = 1  # to tell is a starting balance
                imported_transactions.append(reconciled_transaction)
        # Consume transactions in the ascending order
        for transaction in new_transactions[::-1]:
            if not transaction.booked:
                continue
            payee = transaction.payee_name or ""
            reconciled = reconcile_transaction(
                self.session,
                transaction.date,
                acct,
                payee,
                transaction.notes,
                amount=transaction.transaction_amount.amount,
                imported_id=transaction.transaction_id,
                cleared=transaction.booked,
                imported_payee=payee,
                already_matched=imported_transactions,
            )
            if reconciled.changed():
                imported_transactions.append(reconciled)
        return imported_transactions

    def run_bank_sync(
        self, account: str | Accounts | None = None, start_date: datetime.date | None = None, run_rules: bool = False
    ) -> List[Transactions]:
        """
        Runs the bank synchronization for the selected account. If missing, all accounts are synchronized. If a
        start_date is provided, is used as a reference, otherwise, the last timestamp of each account will be used. If
        the account does not have any transaction, the last 90 days are considered instead.

        If the `start_date` is not provided and the account does not have any transaction, a reconcile transaction will
        be generated to match the expected balance of the account. This would correct the account balance with the
        remote one.

        If `run_rules` is set, the rules will be run for the imported transactions. Please note that unlike Actual,
        the rules here are ran at the final imported objects. This is unlikely to cause data mismatches,
        but if you find any issues feel free to report this as an issue.
        """
        # if no account is provided, sync all of them, otherwise just the account provided
        if account is None:
            accounts = get_accounts(self.session)
        else:
            account = get_account(self.session, account)
            accounts = [account]
        imported_transactions = []

        default_start_date: datetime.date = start_date
        is_first_sync: bool = False
        for acct in accounts:
            sync_method = acct.account_sync_source
            account_id = acct.account_id
            if not (account_id and sync_method):
                continue
            status = self.bank_sync_status(sync_method.lower())
            if not status.data.configured:
                continue
            if start_date is None:
                all_transactions = get_transactions(self.session, account=acct)
                if all_transactions:
                    default_start_date = all_transactions[0].get_date()
                else:
                    is_first_sync = True
                    default_start_date = datetime.date.today() - datetime.timedelta(days=90)
            transactions = self._run_bank_sync_account(acct, default_start_date, is_first_sync)
            imported_transactions.extend(transactions)
        if run_rules:
            self.run_rules(imported_transactions)
        return imported_transactions



from __future__ import annotations

import enum
from typing import Dict, List, Literal, Optional, Union

from pydantic import BaseModel, Field, TypeAdapter

from actual.api.bank_sync import (
    BankSyncAccountData,
    BankSyncErrorData,
    BankSyncTransactionData,
)


class Endpoints(enum.Enum):
    LOGIN = "account/login"
    INFO = "info"
    ACCOUNT_VALIDATE = "account/validate"
    NEEDS_BOOTSTRAP = "account/needs-bootstrap"
    BOOTSTRAP = "account/bootstrap"
    LOGIN_METHODS = "account/login-methods"
    RESET_PASSWORD = "account/change-password"
    SYNC = "sync/sync"
    LIST_USER_FILES = "sync/list-user-files"
    GET_USER_FILE_INFO = "sync/get-user-file-info"
    UPDATE_USER_FILE_NAME = "sync/update-user-filename"
    DOWNLOAD_USER_FILE = "sync/download-user-file"
    UPLOAD_USER_FILE = "sync/upload-user-file"
    RESET_USER_FILE = "sync/reset-user-file"
    DELETE_USER_FILE = "sync/delete-user-file"
    # encryption related
    USER_GET_KEY = "sync/user-get-key"
    USER_CREATE_KEY = "sync/user-create-key"
    # data related
    DATA_FILE_INDEX = "data-file-index.txt"
    DEFAULT_DB = "data/default-db.sqlite"
    MIGRATIONS = "data/migrations"
    # bank sync related
    SECRET = "secret"
    BANK_SYNC_STATUS = "{bank_sync}/status"
    BANK_SYNC_ACCOUNTS = "{bank_sync}/accounts"
    BANK_SYNC_TRANSACTIONS = "{bank_sync}/transactions"
    # OpenID related
    OPEN_ID_OWNER_CREATED = "admin/owner-created/"  # returns a bool, no model required
    OPEN_ID_CONFIG = "openid/config"
    OPEN_ID_USERS = "admin/users"
    OPEN_ID_ACCESS_USERS = "admin/access/users"
    OPEN_ID_ENABLE = "openid/enable"
    OPEN_ID_DISABLE = "openid/disable"

    def __str__(self):
        return self.value


class BankSyncs(enum.Enum):
    GOCARDLESS = "gocardless"
    SIMPLEFIN = "simplefin"


class StatusCode(enum.Enum):
    OK = "ok"
    ERROR = "error"


class StatusDTO(BaseModel):
    status: StatusCode


class ErrorStatusDTO(BaseModel):
    status: StatusCode
    reason: Optional[str] = None


class TokenDTO(BaseModel):
    """Here, if you try to log in with a password, you will get a token, and if you try to log in with an OpenID,
    you will get a return_url."""

    token: Optional[str] = None
    return_url: Optional[str] = Field(None, alias="returnUrl")


class LoginDTO(StatusDTO):
    data: TokenDTO


class UploadUserFileDTO(StatusDTO):
    group_id: str = Field(..., alias="groupId")


class IsValidatedDTO(BaseModel):
    validated: Optional[bool]
    # optional OpenID fields
    user_name: Optional[str] = Field(None, alias="userName")
    permission: Optional[str] = None
    user_id: Optional[str] = Field(None, alias="userId")
    display_name: Optional[str] = Field(None, alias="displayName")
    login_method: Optional[str] = Field(default="password", alias="loginMethod")


class ValidateDTO(StatusDTO):
    data: IsValidatedDTO


class EncryptMetaDTO(BaseModel):
    key_id: Optional[str] = Field(..., alias="keyId")
    algorithm: Optional[str]
    iv: Optional[str]
    auth_tag: Optional[str] = Field(..., alias="authTag")


class EncryptionTestDTO(BaseModel):
    value: str
    meta: EncryptMetaDTO


class EncryptionDTO(BaseModel):
    id: Optional[str]
    salt: Optional[str]
    test: Optional[str]

    def meta(self) -> EncryptionTestDTO:
        return EncryptionTestDTO.parse_raw(self.test)


class FileDTO(BaseModel):
    deleted: Optional[int]
    file_id: Optional[str] = Field(..., alias="fileId")
    group_id: Optional[str] = Field(..., alias="groupId")
    name: Optional[str]


class RemoteFileListDTO(FileDTO):
    encrypt_key_id: Optional[str] = Field(..., alias="encryptKeyId")
    # optional OpenId fields
    owner: Optional[str] = None
    users_with_access: Optional[List[BaseOpenIDUserFileAccessDTO]] = Field(
        default_factory=list, alias="usersWithAccess"
    )


class RemoteFileDTO(FileDTO):
    encrypt_meta: Optional[EncryptMetaDTO] = Field(..., alias="encryptMeta")


class GetUserFileInfoDTO(StatusDTO):
    data: RemoteFileDTO


class ListUserFilesDTO(StatusDTO):
    data: List[RemoteFileListDTO]


class UserGetKeyDTO(StatusDTO):
    data: EncryptionDTO


class BuildDTO(BaseModel):
    name: str
    description: Optional[str]
    version: Optional[str]


class InfoDTO(BaseModel):
    build: BuildDTO


class LoginMethodDTO(BaseModel):
    method: str
    active: bool
    display_name: str = Field(..., alias="displayName")


class IsBootstrapedDTO(BaseModel):
    bootstrapped: bool
    login_method: Optional[str] = Field(default="password", alias="loginMethod")
    multi_user: Optional[bool] = Field(default=False, alias="multiuser")
    available_login_methods: Optional[List[LoginMethodDTO]] = Field(default=None, alias="availableLoginMethods")


class LoginMethodsDTO(StatusDTO):
    methods: List[LoginMethodDTO]


class BootstrapInfoDTO(StatusDTO):
    data: IsBootstrapedDTO


class IsConfiguredDTO(BaseModel):
    configured: bool


class BankSyncStatusDTO(StatusDTO):
    data: IsConfiguredDTO


class BankSyncAccountDTO(StatusDTO):
    data: BankSyncAccountData


class BankSyncTransactionResponseDTO(StatusDTO):
    data: BankSyncTransactionData


class BankSyncErrorDTO(StatusDTO):
    data: BankSyncErrorData


class IssuerConfig(BaseModel):
    name: str = Field(..., description="Friendly name for the issuer")
    authorization_endpoint: str = Field(..., description="Authorization endpoint URL")
    token_endpoint: str = Field(..., description="Token endpoint URL")
    userinfo_endpoint: str = Field(..., description="User info endpoint URL")


class OpenIDConfigDTO(BaseModel):
    doc: str = Field(default="OpenID authentication settings.", description="Documentation string")
    discovery_url: Optional[str] = Field(alias="discoveryURL")
    issuer: Optional[IssuerConfig]
    client_id: str
    client_secret: str
    server_hostname: str
    auth_method: Literal["openid", "oauth2"] = Field(alias="authMethod")


class OpenIDConfigResponseDTO(StatusDTO):
    data: Dict[str, OpenIDConfigDTO]


class OpenIDBootstrapDTO(BaseModel):
    client_id: str = Field(..., description="OAuth2 client ID")
    client_secret: str = Field(..., description="OAuth2 client secret")
    discovery_url: Optional[IssuerConfig] = Field(
        default=None, alias="discoveryURL", description="OpenID discovery URL"
    )
    server_hostname: str


class OpenIDUserDTO(BaseModel):
    id: str
    user_name: str = Field(..., alias="userName")
    display_name: Optional[str] = Field(..., alias="displayName")
    enabled: bool
    owner: bool
    role: Optional[str] = Field(..., description="User role (ADMIN or BASIC)")


class OpenIDDeleteUserDTO(BaseModel):
    some_deletions_failed: bool = Field(..., alias="someDeletionsFailed")


class OpenIDDeleteUserResponseDTO(StatusDTO):
    data: OpenIDDeleteUserDTO


class BaseOpenIDUserFileAccessDTO(BaseModel):
    user_id: str = Field(..., alias="userId")
    user_name: str = Field(..., alias="userName")
    display_name: Optional[str] = Field(..., alias="displayName")
    owner: bool


class OpenIDUserFileAccessDTO(BaseOpenIDUserFileAccessDTO):
    have_access: bool = Field(..., alias="haveAccess")


BankSyncAccountResponseDTO = TypeAdapter(Union[BankSyncErrorDTO, BankSyncAccountDTO])
BankSyncResponseDTO = TypeAdapter(Union[BankSyncErrorDTO, BankSyncTransactionResponseDTO])



import datetime
import json
import pathlib
from typing import List

import pytest
from click.testing import Result
from testcontainers.core.container import DockerContainer
from testcontainers.core.waiting_utils import wait_for_logs
from typer.testing import CliRunner

from actual import Actual, __version__
from actual.cli.config import Config, default_config_path
from actual.queries import create_account, create_transaction

runner = CliRunner()
server_version = "25.4.0"


def base_dataset(actual: Actual, budget_name: str = "Test", encryption_password: str = None):
    actual.create_budget(budget_name)
    bank = create_account(actual.session, "Bank")
    create_transaction(
        actual.session, datetime.date(2024, 9, 5), bank, "Starting Balance", category="Starting", amount=150
    )
    create_transaction(
        actual.session, datetime.date(2024, 12, 24), bank, "Shopping Center", "Christmas Gifts", "Gifts", -100
    )
    actual.commit()
    actual.upload_budget()
    if encryption_password:
        actual.encrypt(encryption_password)


@pytest.fixture(scope="module")
def actual_server(request, module_mocker, tmp_path_factory):
    path = pathlib.Path(tmp_path_factory.mktemp("config"))
    module_mocker.patch("actual.cli.config.default_config_path", return_value=path / "config.yaml")
    with DockerContainer(f"actualbudget/actual-server:{server_version}").with_exposed_ports(5006) as container:
        wait_for_logs(container, "Listening on :::5006...")
        # create a new budget
        port = container.get_exposed_port(5006)
        with Actual(f"http://localhost:{port}", password="mypass", bootstrap=True) as actual:
            base_dataset(actual)
        # init configuration
        result = invoke(
            [
                "init",
                "--url",
                f"http://localhost:{port}",
                "--password",
                "mypass",
                "--file",
                "Test",
                "--context",
                "test",
            ]
        )
        assert result.exit_code == 0
        yield container


def invoke(command: List[str]) -> Result:
    from actual.cli.main import app

    return runner.invoke(app, command)


def test_init_interactive(actual_server, mocker):
    # create a new encrypted file
    port = actual_server.get_exposed_port(5006)
    with Actual(f"http://localhost:{port}", password="mypass") as actual:
        base_dataset(actual, "Extra", "mypass")
    # test full prompt
    mock_prompt = mocker.patch("typer.prompt")
    mock_prompt.side_effect = [f"http://localhost:{port}", "mypass", 2, "mypass", "myextra"]
    assert invoke(["init"]).exit_code == 0
    assert invoke(["use-context", "myextra"]).exit_code == 0
    assert invoke(["use-context", "test"]).exit_code == 0
    # remove extra context
    assert invoke(["remove-context", "myextra"]).exit_code == 0
    # different context should not succeed
    assert invoke(["use-context", "myextra"]).exit_code != 0
    assert invoke(["remove-context", "myextra"]).exit_code != 0


def test_load_config(actual_server):
    cfg = Config.load()
    assert cfg.default_context == "test"
    assert str(default_config_path()).endswith(".actualpy/config.yaml")
    # if the context does not exist, it should fail to load the server
    cfg.default_context = "foo"
    with pytest.raises(ValueError, match="Could not find budget with context"):
        cfg.actual()


def test_app(actual_server):
    result = invoke(["version"])
    assert result.exit_code == 0
    assert result.stdout == f"Library Version: {__version__}\nServer Version: {server_version}\n"
    # make sure json is valid
    result = invoke(["-o", "json", "version"])
    assert json.loads(result.stdout) == {"library_version": __version__, "server_version": server_version}


def test_metadata(actual_server):
    result = invoke(["metadata"])
    assert result.exit_code == 0
    assert "" in result.stdout
    # make sure json is valid
    result = invoke(["-o", "json", "metadata"])
    assert "budgetName" in json.loads(result.stdout)


def test_accounts(actual_server):
    result = invoke(["accounts"])
    assert result.exit_code == 0
    assert result.stdout == (
        "         Accounts         \n"
        "┏━━━━━━━━━━━━━━┳━━━━━━━━━┓\n"
        "┃ Account Name ┃ Balance ┃\n"
        "┡━━━━━━━━━━━━━━╇━━━━━━━━━┩\n"
        "│ Bank         │   50.00 │\n"
        "└──────────────┴─────────┘\n"
    )
    # make sure json is valid
    result = invoke(["-o", "json", "accounts"])
    assert json.loads(result.stdout) == [{"name": "Bank", "balance": 50.00}]


def test_transactions(actual_server):
    result = invoke(["transactions"])
    assert result.exit_code == 0
    assert result.stdout == (
        "                              Transactions                              \n"
        "┏━━━━━━━━━━━━┳━━━━━━━━━━━━━━━━━━┳━━━━━━━━━━━━━━━━━┳━━━━━━━━━━┳━━━━━━━━━┓\n"
        "┃ Date       ┃ Payee            ┃ Notes           ┃ Category ┃  Amount ┃\n"
        "┡━━━━━━━━━━━━╇━━━━━━━━━━━━━━━━━━╇━━━━━━━━━━━━━━━━━╇━━━━━━━━━━╇━━━━━━━━━┩\n"
        "│ 2024-12-24 │ Shopping Center  │ Christmas Gifts │ Gifts    │ -100.00 │\n"
        "│ 2024-09-05 │ Starting Balance │                 │ Starting │  150.00 │\n"
        "└────────────┴──────────────────┴─────────────────┴──────────┴─────────┘\n"
    )
    # make sure json is valid
    result = invoke(["-o", "json", "transactions"])
    assert {
        "date": "2024-12-24",
        "payee": "Shopping Center",
        "notes": "Christmas Gifts",
        "category": "Gifts",
        "amount": -100.00,
    } in json.loads(result.stdout)


def test_payees(actual_server):
    result = invoke(["payees"])
    assert result.exit_code == 0
    assert result.stdout == (
        "            Payees            \n"
        "┏━━━━━━━━━━━━━━━━━━┳━━━━━━━━━┓\n"
        "┃ Name             ┃ Balance ┃\n"
        "┡━━━━━━━━━━━━━━━━━━╇━━━━━━━━━┩\n"
        "│                  │    0.00 │\n"  # this is the payee for the account
        "│ Starting Balance │  150.00 │\n"
        "│ Shopping Center  │ -100.00 │\n"
        "└──────────────────┴─────────┘\n"
    )
    # make sure json is valid
    result = invoke(["-o", "json", "payees"])
    assert {"name": "Shopping Center", "balance": -100.00} in json.loads(result.stdout)


def test_export(actual_server, mocker):
    export_data = mocker.patch("actual.Actual.export_data")
    invoke(["export"])
    export_data.assert_called_once()
    assert export_data.call_args[0][0].name.endswith("Test.zip")

    # test normal file name
    invoke(["export", "Test.zip"])
    assert export_data.call_count == 2
    assert export_data.call_args[0][0].name == "Test.zip"



from __future__ import annotations

import datetime
import decimal
import enum
from typing import List, Optional

from pydantic import AliasChoices, BaseModel, Field

from actual.utils.title import title


class BankSyncTransactionDTO(BaseModel):
    id: str
    posted: int
    amount: str
    description: str
    payee: str
    memo: str


class BankSyncOrgDTO(BaseModel):
    domain: str
    sfin_url: str = Field(..., alias="sfin-url")


class BankSyncAccountDTO(BaseModel):
    org: BankSyncOrgDTO
    id: str
    name: str
    currency: str
    balance: str
    available_balance: str = Field(..., alias="available-balance")
    balance_date: int = Field(..., alias="balance-date")
    transactions: List[BankSyncTransactionDTO]
    holdings: List[dict]


class BankSyncAmount(BaseModel):
    amount: decimal.Decimal
    currency: str


class DebtorAccount(BaseModel):
    iban: str

    @property
    def masked_iban(self):
        return f"({self.iban[:4]} XXX {self.iban[-4:]})"


class BalanceType(enum.Enum):
    # See https://developer.gocardless.com/bank-account-data/balance#balance_type for full documentation
    CLOSING_AVAILABLE = "closingAvailable"
    CLOSING_BOOKED = "closingBooked"
    CLOSING_CLEARED = "closingCleared"
    EXPECTED = "expected"
    FORWARD_AVAILABLE = "forwardAvailable"
    INTERIM_AVAILABLE = "interimAvailable"
    INTERIM_CLEARED = "interimCleared"
    INFORMATION = "information"
    INTERIM_BOOKED = "interimBooked"
    NON_INVOICED = "nonInvoiced"
    OPENING_BOOKED = "openingBooked"
    OPENING_AVAILABLE = "openingAvailable"
    OPENING_CLEARED = "openingCleared"
    PREVIOUSLY_CLOSED_BOOKED = "previouslyClosedBooked"


class Balance(BaseModel):
    """An object containing the balance amount and currency."""

    balance_amount: BankSyncAmount = Field(..., alias="balanceAmount")
    balance_type: BalanceType = Field(..., alias="balanceType")
    reference_date: Optional[str] = Field(None, alias="referenceDate", description="The date of the balance")


class TransactionItem(BaseModel):
    transaction_id: Optional[str] = Field(None, alias="transactionId")
    booked: Optional[bool] = False
    transaction_amount: BankSyncAmount = Field(..., alias="transactionAmount")
    # these fields are generated on the server itself, so we can trust them as being correct
    payee_name: Optional[str] = Field(None, alias="payeeName")
    date: datetime.date = Field(..., alias="date")
    notes: Optional[str] = Field(None, alias="notes")
    # goCardless optional fields
    payee: Optional[str] = Field(None, validation_alias=AliasChoices("debtorName", "creditorName"))
    payee_account: Optional[DebtorAccount] = Field(
        None, validation_alias=AliasChoices("debtorAccount", "creditorAccount")
    )
    booking_date: Optional[datetime.date] = Field(None, alias="bookingDate")
    value_date: Optional[datetime.date] = Field(None, alias="valueDate")
    remittance_information_unstructured: str = Field(None, alias="remittanceInformationUnstructured")
    remittance_information_unstructured_array: List[str] = Field(
        default_factory=list, alias="remittanceInformationUnstructuredArray"
    )
    additional_information: Optional[str] = Field(None, alias="additionalInformation")
    # simpleFin optional fields
    posted_date: Optional[datetime.date] = Field(None, alias="postedDate")

    @property
    def imported_payee(self):
        """Deprecated method to convert the payee name. Use the payee_name instead."""
        name_parts = []
        name = self.payee or self.notes or self.additional_information
        if name:
            name_parts.append(title(name))
        if self.payee_account and self.payee_account.iban:
            name_parts.append(self.payee_account.masked_iban)
        return " ".join(name_parts).strip()


class Transactions(BaseModel):
    all: List[TransactionItem] = Field(..., description="List of all transactions, from newest to oldest.")
    booked: List[TransactionItem]
    pending: List[TransactionItem]


class BankSyncAccountData(BaseModel):
    accounts: List[BankSyncAccountDTO]


class BankSyncTransactionData(BaseModel):
    balances: List[Balance]
    starting_balance: int = Field(..., alias="startingBalance")
    transactions: Transactions
    # goCardless specific
    iban: Optional[str] = None
    institution_id: Optional[str] = Field(None, alias="institutionId")

    @property
    def balance(self) -> decimal.Decimal:
        """Starting balance of the account integration, converted to a decimal amount.

        For `simpleFin`, this will represent the current amount on the account, while for `goCardless` it will
        represent the actual initial amount before all transactions.
        """
        return decimal.Decimal(self.starting_balance) / 100


class BankSyncErrorData(BaseModel):
    error_type: str
    error_code: str
    status: Optional[str] = None
    reason: Optional[str] = None



import threading
import time

import pytest
from requests import Session, get
from testcontainers.core.container import DockerContainer
from testcontainers.core.waiting_utils import wait_for_logs

from actual import Actual
from actual.exceptions import ActualInvalidOperationError, AuthorizationError
from tests.conftest import RequestsMock


@pytest.fixture()
def actual_server(request):
    # we test integration with the 5 latest versions of actual server
    with DockerContainer("actualbudget/actual-server:25.7.1").with_exposed_ports(5006) as container:
        wait_for_logs(container, "Listening on :::5006...")
        yield container


def test_openid_endpoints(actual_server, mocker):
    port = actual_server.get_exposed_port(5006)
    with Actual(f"http://localhost:{port}", password="mypass", bootstrap=True) as actual:
        actual.create_budget("My Budget")
        actual.upload_budget()
        assert actual.is_open_id_owner_created() is False
        methods = actual.login_methods()
        assert len(methods.methods) == 1
        assert methods.methods[0].method == "password"
        assert actual.open_id_users() == []
        # somehow those are not really validating anything, so we can test those endpoints
        user = actual.create_open_id_user("foo")
        actual.update_open_id_user(user.id, display_name="foobar", owner=False, enabled=True)
        with pytest.raises(ActualInvalidOperationError):
            actual.update_open_id_user("not_existing", display_name="foobar")
        users = actual.open_id_users()
        assert len(users) == 1
        assert user.id == users[0].id
        assert users[0].display_name == "foobar"
        # get permissions of file per user
        permissions = actual.list_file_users_allowed(actual._file.file_id)
        assert len(permissions) == 1
        assert all(user.owner is False for user in permissions)
        # Delete user does not work due to some internal exception (when not set), so we mock the response for now
        mocker.patch.object(Session, "delete").return_value = RequestsMock(
            {"status": "ok", "data": {"someDeletionsFailed": False}}
        )
        actual.delete_open_id_user(user.id)


def test_login_handshake(mocker):

    def _threading_call(url: str):
        # This thread will do the interaction of the user logging in via browser
        # We just wait a second then call the endpoint passing the token from the open id callback to the API
        time.sleep(1)
        get(url, params={"token": "mytoken"})

    def _login_fn(_url: str, json: dict):
        assert "returnUrl" in json
        url = json.get("returnUrl")
        threading.Thread(target=_threading_call, args=(url,)).start()
        return RequestsMock({"status": "ok", "data": {}})

    mocker.patch.object(Actual, "validate")
    mocker.patch.object(Actual, "is_open_id_owner_created", return_value=True)
    mocker.patch.object(Actual, "needs_bootstrap", return_value=True)
    mocker.patch.object(Session, "post").side_effect = _login_fn

    # If the handshake is successful, the token would be set
    with Actual("http://localhost:123") as actual:
        assert actual._token == "mytoken"


def test_login_exceptions(mocker):
    mocker.patch.object(Actual, "validate")
    mocker.patch.object(Actual, "is_open_id_owner_created", return_value=False)

    with pytest.raises(ValueError, match="provide a valid token or a password"):
        with Actual("http://localhost:123"):
            pass
    with pytest.raises(AuthorizationError, match="OpenID server is not set-up"):
        actual = Actual("http://localhost:123", token="foo")
        actual.login(None, "openid")
